from argparse import ArgumentParser
from pathlib import Path
from os import path

import pickle
from types import ModuleType

import torch


class RenameUnpickler(pickle.Unpickler):
    def find_class(self, module, name):
        renamed_module = module
        if module == "pytorch_lightning.utilities.argparse_utils":
            renamed_module = "pytorch_lightning.utilities.argparse"

        return super(RenameUnpickler, self).find_class(renamed_module, name)


class RenamePickleModule(ModuleType):
    def __init__(self):
        super().__init__('pickle')

    def __getattr__(self, item):
        if item == 'Unpickler':
            return RenameUnpickler

        return getattr(pickle, item)


def patch(ckpt_path, patched_path=None):
    if patched_path is None:
        ckpt_path_utils = Path(ckpt_path)
        patched_path = path.join(ckpt_path_utils.parent, f"patched_{ckpt_path_utils.name}")

    if path.exists(patched_path):
        return patched_path
    ckpt = torch.load(ckpt_path, map_location='cpu', pickle_module=RenamePickleModule())
    torch.save(ckpt, patched_path)
    return patched_path


def main():
    parser = ArgumentParser()
    parser.add_argument("--ckpt", type=str, required=True)
    parser.add_argument("--patched", type=str, default=None)
    args = parser.parse_args()

    patch(args.ckpt, args.patched)


if __name__ == '__main__':
    main()



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

import re
import inspect
from argparse import ArgumentParser

from transformers import optimization
from transformers.optimization import *


def get_layer_lrs(
        named_parameters,
        transformer_preffix,
        learning_rate,
        layer_decay,
        n_layers,
        **kwargs
):
    groups = []
    temp_groups = [None] * (n_layers + 3)
    temp_no_decay_groups = [None] * (n_layers + 3)
    for name, parameters in named_parameters:
        regex = rf"^{transformer_preffix}\.(embeddings|encoder)\w*\.(layer.(\d+))?.+"
        m = re.match(regex, name)

        is_transformer = True
        if m is None:
            depth = n_layers + 2
            is_transformer = False
        elif m.group(1) == 'embeddings':
            depth = 0
        elif m.group(1) == 'encoder':
            depth = int(m.group(3)) + 1
        else:
            raise Exception("Not Recommand!!!")

        if is_transformer and any(x in name for x in ['bias', 'LayerNorm.bias', 'LayerNorm.weight']):
            if temp_no_decay_groups[depth] is None:
                temp_no_decay_groups[depth] = []
            temp_no_decay_groups[depth].append(parameters)
        else:
            if temp_groups[depth] is None:
                temp_groups[depth] = []
            temp_groups[depth].append(parameters)

    for depth, parameters in enumerate(temp_no_decay_groups):
        if parameters:
            groups.append({
                'params': parameters, 'weight_decay': 0.0, 'lr': learning_rate * (layer_decay ** (n_layers + 2 - depth))
            })
    for depth, parameters in enumerate(temp_groups):
        if parameters:
            groups.append({
                'params': parameters, 'lr': learning_rate * (layer_decay ** (n_layers + 2 - depth))
            })
    return groups


def get_layer_lrs_with_crf(
        named_parameters,
        transformer_preffix,
        learning_rate,
        layer_decay,
        n_layers,
        crf_preffix,
        crf_rate=10.0,
        **kwargs
):
    groups = []
    crf_groups = []
    temp_groups = [None] * (n_layers + 3)
    temp_no_decay_groups = [None] * (n_layers + 3)
    for name, parameters in named_parameters:
        regex = rf"^{transformer_preffix}\.(embeddings|encoder)\w*\.(layer.(\d+))?.+"
        m = re.match(regex, name)

        is_transformer = True
        if m is None:
            depth = n_layers + 2
            is_transformer = False
        elif m.group(1) == 'embeddings':
            depth = 0
        elif m.group(1) == 'encoder':
            depth = int(m.group(3)) + 1
        else:
            raise Exception("Not Recommand!!!")

        if is_transformer and any(x in name for x in ['bias', 'LayerNorm.bias', 'LayerNorm.weight']):
            if temp_no_decay_groups[depth] is None:
                temp_no_decay_groups[depth] = []
            temp_no_decay_groups[depth].append(parameters)
        elif not is_transformer and crf_preffix in name:
            crf_groups.append(parameters)
        else:
            if temp_groups[depth] is None:
                temp_groups[depth] = []
            temp_groups[depth].append(parameters)

    for depth, parameters in enumerate(temp_no_decay_groups):
        if parameters:
            groups.append({
                'params': parameters, 'weight_decay': 0.0, 'lr': learning_rate * (layer_decay ** (n_layers + 2 - depth))
            })
    for depth, parameters in enumerate(temp_groups):
        if parameters:
            groups.append({
                'params': parameters, 'lr': learning_rate * (layer_decay ** (n_layers + 2 - depth))
            })
    if crf_groups:
        groups.append({
            'params': crf_groups, 'lr': learning_rate * crf_rate
        })

    return groups


scheduler_register = {
    'constant_schedule': get_constant_schedule,
    'constant_schedule_with_warmup': get_constant_schedule_with_warmup,
    'linear_schedule_with_warmup': get_linear_schedule_with_warmup,
    'cosine_schedule_with_warmup': get_cosine_schedule_with_warmup,
    'cosine_with_hard_restarts_schedule_with_warmup': get_cosine_with_hard_restarts_schedule_with_warmup,
    'polynomial_decay_schedule_with_warmup': get_polynomial_decay_schedule_with_warmup
}


def get_scheduler(optimizer, name, **kwargs) -> LambdaLR:
    scheduler = scheduler_register.get(name, get_linear_schedule_with_warmup)
    signature = inspect.signature(scheduler)
    kwargs = {key: value for key, value in kwargs.items() if key in signature.parameters}
    return scheduler(optimizer, **kwargs)


layer_lrs_getters_register = {
    'get_layer_lrs': get_layer_lrs,
    'get_layer_lrs_with_crf': get_layer_lrs_with_crf,
}


def get_layer_lrs_getters(named_parameters, name, **kwargs):
    layer_lrs_getter = layer_lrs_getters_register.get(name, get_layer_lrs)
    signature = inspect.signature(layer_lrs_getter)
    kwargs = {key: value for key, value in kwargs.items() if key in signature.parameters}
    return layer_lrs_getter(named_parameters, **kwargs)


def create_optimizer(
        model, lr,
        num_train_steps,
        beta1=0.9,
        beta2=0.999,
        weight_decay=0.0,
        warmup_steps=0,
        warmup_proportion=0.1,
        layerwise_lr_decay_power=0.8,
        transformer_preffix="transformer",
        n_transformer_layers=12,
        layer_lrs_getter='get_layer_lrs',
        layer_lrs_getter_kwargs=None,
        lr_scheduler='linear_schedule_with_warmup',
        lr_scheduler_kwargs=None,
):
    """

    Args:
        model:
        lr: 3e-4 for Small, 1e-4 for Base, 5e-5 for Large
        num_train_steps:
        weight_decay: 0
        warmup_steps: 0
        warmup_proportion: 0.1
        lr_decay_power: 1.0
        layerwise_lr_decay_power: 0.8 for Base/Small, 0.9 for Large

    Returns:

    """
    if lr_scheduler_kwargs is None:
        lr_scheduler_kwargs = {}
    if layer_lrs_getter_kwargs is None:
        layer_lrs_getter_kwargs = {}
    if lr_scheduler_kwargs is None:
        lr_scheduler_kwargs = {}
    if layerwise_lr_decay_power > 0:
        parameters = get_layer_lrs_getters(
            named_parameters=list(model.named_parameters()),
            name=layer_lrs_getter,
            transformer_preffix=transformer_preffix,
            learning_rate=lr,
            layer_decay=layerwise_lr_decay_power,
            n_layers=n_transformer_layers,
            **layer_lrs_getter_kwargs
        )
    else:
        parameters = model.parameters()

    optimizer = optimization.AdamW(
        parameters,
        lr=lr,
        weight_decay=weight_decay,
        betas=(beta1, beta2),
        eps=1e-6,
        correct_bias=False,
    )

    warmup_steps = max(num_train_steps * warmup_proportion, warmup_steps)
    scheduler = get_scheduler(
        optimizer,
        name=lr_scheduler,
        num_warmup_steps=warmup_steps,
        num_training_steps=num_train_steps,
        **lr_scheduler_kwargs
    )
    return optimizer, scheduler


def add_optimizer_specific_args(parent_parser):
    parser = ArgumentParser(parents=[parent_parser], add_help=False)
    parser.add_argument('--beta1', type=float, default=0.9)
    parser.add_argument('--beta2', type=float, default=0.999)

    # 3e-4 for Small, 1e-4 for Base, 5e-5 for Large
    parser.add_argument('--lr', type=float, default=1e-4)

    parser.add_argument(
        '--lr_scheduler', type=str, default='linear_schedule_with_warmup', choices=scheduler_register.keys()
    )
    parser.add_argument('--lr_end', type=float, default=1e-7)
    parser.add_argument('--lr_num_cycles', type=float, default=0.5)
    parser.add_argument('--lr_decay_power', type=float, default=1.0)

    parser.add_argument(
        '--lr_layers_getter', type=str, default='get_layer_lrs_with_crf', choices=layer_lrs_getters_register.keys()
    )
    parser.add_argument('--lr_crf_preffix', type=str, default='crf')
    parser.add_argument('--lr_crf_rate', type=float, default=10.0)

    parser.add_argument('--warmup_steps', type=int, default=0)
    parser.add_argument('--warmup_proportion', type=float, default=0.1)
    parser.add_argument('--weight_decay', type=float, default=0.0)

    # 0.8 for Base/Small, 0.9 for Large
    parser.add_argument('--layerwise_lr_decay_power', type=float, default=0.8)
    return parser


def from_argparse_args(args, model, num_train_steps, n_transformer_layers=12, **kwargs):
    return create_optimizer(
        model,
        lr=args.lr,
        beta1=args.beta1,
        beta2=args.beta2,
        num_train_steps=num_train_steps,
        weight_decay=args.weight_decay,
        warmup_steps=args.warmup_steps,
        warmup_proportion=args.warmup_proportion,
        layerwise_lr_decay_power=args.layerwise_lr_decay_power,
        n_transformer_layers=n_transformer_layers,
        layer_lrs_getter=args.lr_layers_getter,
        layer_lrs_getter_kwargs={
            'crf_preffix': args.lr_crf_preffix,
            'crf_rate': args.lr_crf_rate,
        },
        lr_scheduler=args.lr_scheduler,
        lr_scheduler_kwargs={
            'lr_end': args.lr_end,
            'power': args.lr_decay_power,
            'num_cycles': args.lr_num_cycles
        }, **kwargs
    )



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

import os
import types
import numpy as np
from argparse import ArgumentParser

import torch
import torch.utils.data
import torch.nn.functional as F

from pytorch_lightning import Trainer

import ltp
from ltp import optimization, multitask
from ltp.data import dataset as datasets
from ltp.data.utils import collate, MultiTaskDataloader
from ltp.transformer_multitask import TransformerMultiTask as Model
from ltp.utils import TaskInfo, common_train, deploy_model, tune_train, map2device, add_common_specific_args
from ltp.utils import add_tune_specific_args
from ltp.multitask import validation_method

os.environ['TOKENIZERS_PARALLELISM'] = 'true'


def kd_ce_loss(logits_S, logits_T, temperature=1):
    '''
    Calculate the cross entropy between logits_S and logits_T

    :param logits_S: Tensor of shape (batch_size, length, num_labels) or (batch_size, num_labels)
    :param logits_T: Tensor of shape (batch_size, length, num_labels) or (batch_size, num_labels)
    :param temperature: A float or a tensor of shape (batch_size, length) or (batch_size,)
    '''
    if isinstance(temperature, torch.Tensor) and temperature.dim() > 0:
        temperature = temperature.unsqueeze(-1)
    beta_logits_T = logits_T / temperature
    beta_logits_S = logits_S / temperature
    p_T = F.softmax(beta_logits_T, dim=-1)
    loss = -(p_T * F.log_softmax(beta_logits_S, dim=-1))
    return (temperature * temperature * loss).sum(dim=-1).mean()


def kd_mse_loss(logits_S, logits_T, temperature=1):
    '''
    Calculate the mse loss between logits_S and logits_T

    :param logits_S: Tensor of shape (batch_size, length, num_labels) or (batch_size, num_labels)
    :param logits_T: Tensor of shape (batch_size, length, num_labels) or (batch_size, num_labels)
    :param temperature: A float or a tensor of shape (batch_size, length) or (batch_size,)
    '''
    if isinstance(temperature, torch.Tensor) and temperature.dim() > 0:
        temperature = temperature.unsqueeze(-1)
    beta_logits_T = logits_T / temperature
    beta_logits_S = logits_S / temperature
    loss = F.mse_loss(beta_logits_S, beta_logits_T, reduction='none')
    return (temperature * temperature * loss).mean()


def flsw_temperature_scheduler_builder(beta=1, gamma=2, base_temperature=8, eps=1e-4, *args):
    '''
    adapted from arXiv:1911.07471
    '''

    def flsw_temperature_scheduler(logits_S, logits_T):
        v = logits_S.detach()
        t = logits_T.detach()
        with torch.no_grad():
            v = v / (torch.norm(v, dim=-1, keepdim=True) + eps)
            t = t / (torch.norm(t, dim=-1, keepdim=True) + eps)
            w = torch.pow((1 - (v * t).sum(dim=-1)), gamma)
            tau = base_temperature + (w.mean() - w) * beta
        return tau

    return flsw_temperature_scheduler


def distill_linear(batch, result, target, temperature_scheduler, model: Model = None, extra=None) -> torch.Tensor:
    if 'logits_mask' in batch:
        logits_mask = batch['logits_mask']
    elif 'word_attention_mask' in batch:
        logits_mask = batch['word_attention_mask']
    else:
        logits_mask = batch['attention_mask'][:, 2:]
    active_logits = result.logits[logits_mask]
    active_target_logits = target[logits_mask]

    if result.decoded is not None and extra is not None:
        start_transitions = torch.as_tensor(extra['start_transitions'], device=model.device)
        transitions = torch.as_tensor(extra['transitions'], device=model.device)
        end_transitions = torch.as_tensor(extra['end_transitions'], device=model.device)

        temperature = temperature_scheduler(model.srl_classifier.crf.transitions, transitions)
        kd_loss = kd_mse_loss(active_logits, active_target_logits, temperature)

        transitions_temp = temperature_scheduler(model.srl_classifier.crf.transitions, transitions)
        s_transitions_temp = temperature_scheduler(model.srl_classifier.crf.start_transitions, start_transitions)
        e_transitions_temp = temperature_scheduler(model.srl_classifier.crf.end_transitions, end_transitions)

        crf_loss = kd_mse_loss(transitions, model.srl_classifier.crf.transitions, transitions_temp) + \
                   kd_mse_loss(start_transitions, model.srl_classifier.crf.start_transitions, s_transitions_temp) + \
                   kd_mse_loss(end_transitions, model.srl_classifier.crf.end_transitions, e_transitions_temp)

        return kd_loss + crf_loss
    else:
        temperature = temperature_scheduler(active_logits, active_target_logits)
        return kd_ce_loss(active_logits, active_target_logits, temperature=temperature)


def distill_matrix_dep(batch, result, target, temperature_scheduler, model: Model = None, extra=None) -> torch.Tensor:
    head = batch['head']
    logits_mask = batch['word_attention_mask']

    # Only keep active parts of the loss
    active_heads = head[logits_mask]

    arc_logits, rel_logits = result.src_arc_logits, result.rel_logits
    target_arc_logits, target_rel_logits = target

    arc_logits = arc_logits[:, 1:, :][logits_mask]
    target_arc_logits = target_arc_logits[:, 1:, :][logits_mask]

    rel_logits = rel_logits[:, 1:, :][logits_mask][torch.arange(len(active_heads)), active_heads]
    target_rel_logits = target_rel_logits[:, 1:, :][logits_mask][torch.arange(len(active_heads)), active_heads]

    temperature = temperature_scheduler(arc_logits, target_arc_logits)
    arc_loss = kd_ce_loss(arc_logits, target_arc_logits, temperature=temperature)

    temperature = temperature_scheduler(rel_logits, target_rel_logits)
    rel_loss = kd_ce_loss(rel_logits, target_rel_logits, temperature=temperature)

    classifier = model.dep_classifier
    return 2 * ((1 - classifier.loss_interpolation) * arc_loss + classifier.loss_interpolation * rel_loss)


def distill_matrix_sdp(batch, result, target, temperature_scheduler, model: Model = None, extra=None) -> torch.Tensor:
    head = batch['head']
    logits_mask = batch['word_attention_mask']

    arc_logits, rel_logits = result.src_arc_logits, result.rel_logits
    target_arc_logits, target_rel_logits = target

    arc_logits = arc_logits[:, 1:, :][logits_mask]
    target_arc_logits = target_arc_logits[:, 1:, :][logits_mask]

    rel_logits = rel_logits[:, 1:, :][head > 0]
    target_rel_logits = target_rel_logits[:, 1:, :][head > 0]

    temperature = temperature_scheduler(arc_logits, target_arc_logits)
    arc_loss = kd_mse_loss(arc_logits, target_arc_logits, temperature=temperature)

    temperature = temperature_scheduler(rel_logits, target_rel_logits)
    rel_loss = kd_ce_loss(rel_logits, target_rel_logits, temperature=temperature)

    classifier = model.dep_classifier
    return 2 * ((1 - classifier.loss_interpolation) * arc_loss + classifier.loss_interpolation * rel_loss)


def distill_matrix_crf(batch, result, target, temperature_scheduler, model: Model = None, extra=None) -> torch.Tensor:
    if 'word_attention_mask' in batch:
        logits_mask = batch['word_attention_mask']
    else:
        logits_mask = batch['attention_mask'][:, 2:]

    logits_mask = logits_mask.unsqueeze_(-1).expand(-1, -1, logits_mask.size(1))
    logits_mask = logits_mask & logits_mask.transpose(-1, -2)
    logits_mask = logits_mask.flatten(end_dim=1)
    index = logits_mask[:, 0]
    logits_mask = logits_mask[index]

    s_rel, labels = result.rel_logits, result.labels
    t_rel = target

    active_logits = s_rel[logits_mask]
    active_target_logits = t_rel[logits_mask]

    temperature = temperature_scheduler(active_logits, active_target_logits)
    kd_loss = kd_mse_loss(active_logits, active_target_logits, temperature)

    start_transitions = torch.as_tensor(extra['start_transitions'], device=model.device)
    transitions = torch.as_tensor(extra['transitions'], device=model.device)
    end_transitions = torch.as_tensor(extra['end_transitions'], device=model.device)

    transitions_temp = temperature_scheduler(model.srl_classifier.crf.transitions, transitions)
    s_transitions_temp = temperature_scheduler(model.srl_classifier.crf.start_transitions, start_transitions)
    e_transitions_temp = temperature_scheduler(model.srl_classifier.crf.end_transitions, end_transitions)

    crf_loss = kd_mse_loss(transitions, model.srl_classifier.crf.transitions, transitions_temp) + \
               kd_mse_loss(start_transitions, model.srl_classifier.crf.start_transitions, s_transitions_temp) + \
               kd_mse_loss(end_transitions, model.srl_classifier.crf.end_transitions, e_transitions_temp)
    return kd_loss + crf_loss


distill_loss_map = {
    'seg': distill_linear,
    'pos': distill_linear,
    'ner': distill_linear,
    'dep': distill_matrix_dep,
    'sdp': distill_matrix_sdp,
    'srl': distill_matrix_crf,
}


def build_dataset(model, **kwargs):
    kwargs = {key: value for key, value in kwargs.items() if value is not None}

    distill_datasets = {}
    distill_datasets_extra = {}

    for task, task_data_dir in kwargs.items():
        task_distill_path = os.path.join(task_data_dir, task, 'output.npz')
        task_distill_data = np.load(task_distill_path, allow_pickle=True)

        distill_datasets[task] = task_distill_data['data'].tolist()
        distill_datasets_extra[task] = task_distill_data.get('extra', None)
        if distill_datasets_extra[task] is not None:
            distill_datasets_extra[task] = distill_datasets_extra[task].tolist()

    datasets, metrics = multitask.build_dataset(model, **kwargs)
    return (datasets, distill_datasets, distill_datasets_extra), metrics


def build_method(model: Model, task_info: TaskInfo):
    (multi_dataset, distill_datasets, distill_datasets_extra), multi_metric = build_dataset(
        model,
        seg=model.hparams.seg_data_dir,
        pos=model.hparams.pos_data_dir,
        ner=model.hparams.ner_data_dir,
        dep=model.hparams.dep_data_dir,
        sdp=model.hparams.sdp_data_dir,
        srl=model.hparams.srl_data_dir
    )

    disable_distill = {
        'seg': model.hparams.disable_seg,
        'pos': model.hparams.disable_pos,
        'ner': model.hparams.disable_ner,
        'dep': model.hparams.disable_dep,
        'sdp': model.hparams.disable_sdp,
        'srl': model.hparams.disable_srl,
    }

    disable_distill = {task for task, disable in disable_distill.items() if disable}

    temperature_scheduler = flsw_temperature_scheduler_builder(
        beta=model.hparams.distill_beta,
        gamma=model.hparams.distill_gamma,
        base_temperature=model.hparams.temperature
    )

    def train_dataloader(self):
        multi_dataloader = {
            task: torch.utils.data.DataLoader(
                task_dataset,
                batch_size=None,
                num_workers=self.hparams.num_workers,
                pin_memory=True,
                shuffle=True
            )
            for task, task_dataset in distill_datasets.items()
        }
        res = MultiTaskDataloader(tau=self.hparams.tau, **multi_dataloader)
        return res

    def training_step(self: Model, batch, batch_idx):
        task = batch['task']
        target_logits = batch.pop('logits')
        result = self(**batch)
        norm_loss = result.loss

        if task not in disable_distill:
            distill_loss = distill_loss_map[task](
                batch, result, target_logits, temperature_scheduler, model,
                extra=distill_datasets_extra[task]
            )
            distill_loss_weight = self.global_step / self.num_train_steps
            loss = distill_loss_weight * norm_loss + (1 - distill_loss_weight) * distill_loss

            self.log("distill_loss", distill_loss.item())
            self.log("norm_loss", norm_loss.item())
            self.log("loss", loss.item())
            return {"loss": loss}
        else:
            self.log("loss", norm_loss.item())
            return {"loss": norm_loss}

    def val_dataloader(self):
        return [
            torch.utils.data.DataLoader(
                task_dataset[datasets.Split.VALIDATION],
                batch_size=getattr(self.hparams, f'{task}_batch_size') or self.hparams.batch_size,
                collate_fn=collate,
                num_workers=self.hparams.num_workers,
                pin_memory=True
            ) for task, task_dataset in multi_dataset.items()
        ]

    def test_dataloader(self):
        return [
            torch.utils.data.DataLoader(
                task_dataset[datasets.Split.TEST],
                batch_size=getattr(self.hparams, f'{task}_batch_size') or self.hparams.batch_size,
                collate_fn=collate,
                num_workers=self.hparams.num_workers,
                pin_memory=True
            ) for task, task_dataset in multi_dataset.items()
        ]

    # AdamW + LR scheduler
    def configure_optimizers(self: Model):
        num_epoch_steps = sum(len(dataset) for dataset in distill_datasets.values())
        num_train_steps = num_epoch_steps * self.hparams.max_epochs
        setattr(self, 'num_train_steps', num_train_steps)
        optimizer, scheduler = optimization.from_argparse_args(
            self.hparams,
            model=self,
            num_train_steps=num_train_steps,
            n_transformer_layers=self.transformer.config.num_hidden_layers
        )
        return [optimizer], [{'scheduler': scheduler, 'interval': 'step'}]

    model.configure_optimizers = types.MethodType(configure_optimizers, model)

    model.train_dataloader = types.MethodType(train_dataloader, model)
    model.training_step = types.MethodType(training_step, model)

    validation_step, validation_epoch_end = task_info.validation_method(
        multi_metric, task=task_info.task_name, preffix='val'
    )

    model.val_dataloader = types.MethodType(val_dataloader, model)
    model.validation_step = types.MethodType(validation_step, model)
    model.validation_epoch_end = types.MethodType(validation_epoch_end, model)

    test_step, test_epoch_end = task_info.validation_method(
        multi_metric, task=task_info.task_name, preffix='test'
    )

    model.test_dataloader = types.MethodType(test_dataloader, model)
    model.test_step = types.MethodType(test_step, model)
    model.test_epoch_end = types.MethodType(test_epoch_end, model)


task_info = TaskInfo(
    task_name='distill',
    metric_name='metric_mean',
    build_dataset=build_dataset,
    validation_method=validation_method
)


def add_task_specific_args(parent_parser):
    parser = ArgumentParser(parents=[parent_parser], add_help=False)
    parser.add_argument('--disable_seg', action='store_true')
    parser.add_argument('--disable_pos', action='store_true')
    parser.add_argument('--disable_ner', action='store_true')
    parser.add_argument('--disable_srl', action='store_true')
    parser.add_argument('--disable_dep', action='store_true')
    parser.add_argument('--disable_sdp', action='store_true')
    parser.add_argument('--batch_size', type=int, default=24)
    parser.add_argument('--distill_beta', type=float, default=1.0)
    parser.add_argument('--distill_gamma', type=float, default=2.0)
    parser.add_argument('--temperature', type=float, default=8.0)
    parser.add_argument('--num_workers', type=int, default=4)
    parser.add_argument('--tau', type=float, default=0.8)
    parser.add_argument('--ltp_model', type=str, default=None)
    parser.add_argument('--ltp_version', type=str, default=ltp.__version__)
    parser.add_argument('--seg_data_dir', type=str, default=None)
    parser.add_argument('--pos_data_dir', type=str, default=None)
    parser.add_argument('--ner_data_dir', type=str, default=None)
    parser.add_argument('--dep_data_dir', type=str, default=None)
    parser.add_argument('--sdp_data_dir', type=str, default=None)
    parser.add_argument('--srl_data_dir', type=str, default=None)
    parser.add_argument('--seg_batch_size', type=int, default=None)
    parser.add_argument('--pos_batch_size', type=int, default=None)
    parser.add_argument('--ner_batch_size', type=int, default=None)
    parser.add_argument('--dep_batch_size', type=int, default=None)
    parser.add_argument('--sdp_batch_size', type=int, default=None)
    parser.add_argument('--srl_batch_size', type=int, default=None)
    return parser


def main():
    parser = ArgumentParser()

    # add task level args
    parser = add_common_specific_args(parser)
    parser = add_tune_specific_args(parser)
    parser = add_task_specific_args(parser)

    # add model specific args
    parser = Model.add_model_specific_args(parser)
    parser = optimization.add_optimizer_specific_args(parser)
    parser = Trainer.add_argparse_args(parser)

    parser.set_defaults(min_epochs=1, max_epochs=10)
    parser.set_defaults(gradient_clip_val=1.0, lr_layers_getter='get_layer_lrs_with_crf')
    args = parser.parse_args()

    if args.ltp_model is not None and args.resume_from_checkpoint is not None:
        deploy_model(args, args.ltp_version)
    elif args.tune:
        tune_train(args, model_class=Model, task_info=task_info, build_method=build_method)
    else:
        common_train(args, model_class=Model, task_info=task_info, build_method=build_method)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

import torch
import torch.nn as nn
from ltp.nn.dropout import SharedDropout
from torch.nn.modules.rnn import apply_permutation
from torch.nn.utils.rnn import PackedSequence


class LSTM(nn.Module):
    r"""
    LSTM is an variant of the vanilla bidirectional LSTM adopted by Biaffine Parser
    with the only difference of the dropout strategy.
    It drops nodes in the LSTM layers (input and recurrent connections)
    and applies the same dropout mask at every recurrent timesteps.
    APIs are roughly the same as :class:`~torch.nn.LSTM` except that we only allows
    :class:`~torch.nn.utils.rnn.PackedSequence` as input.
    References:
        - Timothy Dozat and Christopher D. Manning. 2017.
          `Deep Biaffine Attention for Neural Dependency Parsing`_.
    Args:
        input_size (int):
            The number of expected features in the input.
        hidden_size (int):
            The number of features in the hidden state `h`.
        num_layers (int):
            The number of recurrent layers. Default: 1.
        bidirectional (bool):
            If ``True``, becomes a bidirectional LSTM. Default: ``False``
        dropout (float):
            If non-zero, introduces a :class:`SharedDropout` layer on the outputs of each LSTM layer except the last layer.
            Default: 0.
    .. _Deep Biaffine Attention for Neural Dependency Parsing:
        https://openreview.net/forum?id=Hk95PK9le
    """

    def __init__(self, input_size, hidden_size, num_layers=1, bidirectional=False, dropout=0):
        super().__init__()

        self.input_size = input_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.bidirectional = bidirectional
        self.dropout = dropout
        self.num_directions = 1 + self.bidirectional

        self.f_cells = nn.ModuleList()
        if bidirectional:
            self.b_cells = nn.ModuleList()
        for _ in range(self.num_layers):
            self.f_cells.append(nn.LSTMCell(input_size=input_size, hidden_size=hidden_size))
            if bidirectional:
                self.b_cells.append(nn.LSTMCell(input_size=input_size, hidden_size=hidden_size))
            input_size = hidden_size * self.num_directions

        self.reset_parameters()

    def __repr__(self):
        s = f"{self.input_size}, {self.hidden_size}"
        if self.num_layers > 1:
            s += f", num_layers={self.num_layers}"
        if self.bidirectional:
            s += f", bidirectional={self.bidirectional}"
        if self.dropout > 0:
            s += f", dropout={self.dropout}"

        return f"{self.__class__.__name__}({s})"

    def reset_parameters(self):
        for param in self.parameters():
            # apply orthogonal_ to weight
            if len(param.shape) > 1:
                nn.init.orthogonal_(param)
            # apply zeros_ to bias
            else:
                nn.init.zeros_(param)

    def permute_hidden(self, hx, permutation):
        if permutation is None:
            return hx
        h = apply_permutation(hx[0], permutation)
        c = apply_permutation(hx[1], permutation)

        return h, c

    def layer_forward(self, x, hx, cell, batch_sizes, reverse=False):
        hx_0 = hx_i = hx
        hx_n, output = [], []
        steps = reversed(range(len(x))) if reverse else range(len(x))
        if self.training:
            hid_mask = SharedDropout.get_mask(hx_0[0], self.dropout)

        for t in steps:
            last_batch_size, batch_size = len(hx_i[0]), batch_sizes[t]
            if last_batch_size < batch_size:
                hx_i = [torch.cat((h, ih[last_batch_size:batch_size])) for h, ih in zip(hx_i, hx_0)]
            else:
                hx_n.append([h[batch_size:] for h in hx_i])
                hx_i = [h[:batch_size] for h in hx_i]
            hx_i = [h for h in cell(x[t], hx_i)]
            output.append(hx_i[0])
            if self.training:
                hx_i[0] = hx_i[0] * hid_mask[:batch_size]
        if reverse:
            hx_n = hx_i
            output.reverse()
        else:
            hx_n.append(hx_i)
            hx_n = [torch.cat(h) for h in zip(*reversed(hx_n))]
        output = torch.cat(output)

        return output, hx_n

    def forward(self, sequence, hx=None):
        r"""
        Args:
            sequence (~torch.nn.utils.rnn.PackedSequence):
                A packed variable length sequence.
            hx (~torch.Tensor, ~torch.Tensor):
                A tuple composed of two tensors `h` and `c`.
                `h` of shape ``[num_layers*num_directions, batch_size, hidden_size]`` holds the initial hidden state
                for each element in the batch.
                `c` of shape ``[num_layers*num_directions, batch_size, hidden_size]`` holds the initial cell state
                for each element in the batch.
                If `hx` is not provided, both `h` and `c` default to zero.
                Default: ``None``.
        Returns:
            ~torch.nn.utils.rnn.PackedSequence, (~torch.Tensor, ~torch.Tensor):
                The first is a packed variable length sequence.
                The second is a tuple of tensors `h` and `c`.
                `h` of shape ``[num_layers*num_directions, batch_size, hidden_size]`` holds the hidden state for `t=seq_len`.
                Like output, the layers can be separated using ``h.view(num_layers, num_directions, batch_size, hidden_size)``
                and similarly for c.
                `c` of shape ``[num_layers*num_directions, batch_size, hidden_size]`` holds the cell state for `t=seq_len`.
        """
        x, batch_sizes = sequence.data, sequence.batch_sizes.tolist()
        batch_size = batch_sizes[0]
        h_n, c_n = [], []

        if hx is None:
            ih = x.new_zeros(self.num_layers * self.num_directions, batch_size, self.hidden_size)
            h, c = ih, ih
        else:
            h, c = self.permute_hidden(hx, sequence.sorted_indices)
        h = h.view(self.num_layers, self.num_directions, batch_size, self.hidden_size)
        c = c.view(self.num_layers, self.num_directions, batch_size, self.hidden_size)

        for i in range(self.num_layers):
            x = torch.split(x, batch_sizes)
            if self.training:
                mask = SharedDropout.get_mask(x[0], self.dropout)
                x = [i * mask[:len(i)] for i in x]
            x_i, (h_i, c_i) = self.layer_forward(x=x,
                                                 hx=(h[i, 0], c[i, 0]),
                                                 cell=self.f_cells[i],
                                                 batch_sizes=batch_sizes)
            if self.bidirectional:
                x_b, (h_b, c_b) = self.layer_forward(x=x,
                                                     hx=(h[i, 1], c[i, 1]),
                                                     cell=self.b_cells[i],
                                                     batch_sizes=batch_sizes,
                                                     reverse=True)
                x_i = torch.cat((x_i, x_b), -1)
                h_i = torch.stack((h_i, h_b))
                c_i = torch.stack((c_i, c_b))
            x = x_i
            h_n.append(h_i)
            c_n.append(c_i)

        x = PackedSequence(x,
                           sequence.batch_sizes,
                           sequence.sorted_indices,
                           sequence.unsorted_indices)
        hx = torch.cat(h_n, 0), torch.cat(c_n, 0)
        hx = self.permute_hidden(hx, sequence.unsorted_indices)

        return x, hx



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

try:
    from pytorch_lightning import LightningModule as BaseModule
except Exception as e:
    from typing import Union, Optional

    from argparse import Namespace
    import torch
    from torch.nn import Module


    class BaseModule(Module):
        __jit_unused_properties__ = ['device', 'dtype']

        def __init__(self):
            super().__init__()
            self._hparams = Namespace()
            self._dtype = torch.get_default_dtype()
            self._device = torch.device('cpu')

        @property
        def dtype(self) -> Union[str, torch.dtype]:
            return self._dtype

        @dtype.setter
        def dtype(self, new_dtype: Union[str, torch.dtype]):
            # necessary to avoid infinite recursion
            raise RuntimeError('Cannot set the dtype explicitly. Please use module.to(new_dtype).')

        @property
        def device(self) -> Union[str, torch.device]:
            return self._device

        @device.setter
        def device(self, new_device: Union[str, torch.device]):
            raise RuntimeError('Cannot set the device explicitly. Please use module.to(new_device).')

        def to(self, *args, **kwargs) -> Module:
            out = torch._C._nn._parse_to(*args, **kwargs)
            self.__update_properties(device=out[0], dtype=out[1])
            return super().to(*args, **kwargs)

        def cuda(self, device: Optional[int] = None) -> Module:
            self.__update_properties(device=torch.device('cuda', index=device))
            return super().cuda(device=device)

        def cpu(self) -> Module:
            self.__update_properties(device=torch.device('cpu'))
            return super().cpu()

        def type(self, dst_type: Union[str, torch.dtype]) -> Module:
            self.__update_properties(dtype=dst_type)
            return super().type(dst_type=dst_type)

        def float(self) -> Module:
            self.__update_properties(dtype=torch.float)
            return super().float()

        def double(self) -> Module:
            self.__update_properties(dtype=torch.double)
            return super().double()

        def half(self) -> Module:
            self.__update_properties(dtype=torch.half)
            return super().half()

        def __update_properties(self, device: Optional[torch.device] = None, dtype: Optional[torch.dtype] = None):

            def apply_fn(module):
                if not isinstance(module, BaseModule):
                    return
                if device is not None:
                    module._device = device
                if dtype is not None:
                    module._dtype = dtype

            self.apply(apply_fn)

        def save_hyperparameters(self, hparams: Namespace):
            self._hparams = Namespace(**hparams.__dict__)

        @property
        def hparams(self):
            return self._hparams



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

import os
import types
from argparse import ArgumentParser
from collections import OrderedDict

import numpy
import torch
import torch.utils.data
from pytorch_lightning import Trainer
from tqdm import tqdm

import ltp
from ltp import (
    optimization,
    task_segmention as seg,
    task_part_of_speech as pos,
    task_named_entity_recognition as ner,
    task_dependency_parsing as dep,
    task_semantic_dependency_parsing as sdp,
    task_semantic_role_labeling as srl
)
from ltp.data import dataset as datasets
from ltp.data.utils import collate, MultiTaskDataloader
from ltp.transformer_multitask import TransformerMultiTask as Model
from ltp.utils import TaskInfo, common_train, tune_train, map2device, convert2npy, add_common_specific_args
from ltp.utils import deploy_model, add_tune_specific_args

os.environ['TOKENIZERS_PARALLELISM'] = 'true'

# CUDA_VISIBLE_DEVICES=0 PYTHONPATH=. python ltp/multitask.py --max_epochs=10 --batch_size=16 --gpus=1 --precision=16 --seg_data_dir=data/seg --pos_data_dir=data/pos --ner_data_dir=data/ner

task_builder = {
    seg.task_info.task_name: seg,
    pos.task_info.task_name: pos,
    ner.task_info.task_name: ner,
    dep.task_info.task_name: dep,
    sdp.task_info.task_name: sdp,
    srl.task_info.task_name: srl,
}


def build_dataset(model, **kwargs):
    kwargs = {key: value for key, value in kwargs.items() if value is not None}

    datasets = OrderedDict()
    metrics = OrderedDict()

    for task, task_data_dir in kwargs.items():
        dataset, metric = task_builder[task].build_dataset(
            data_dir=task_data_dir, task_name=task, model=model
        )
        datasets[task] = dataset
        metrics[task] = metric

    return datasets, metrics


def validation_method(metric: dict = None, task='multi', preffix='val', ret=True):
    if metric is None:
        raise NotImplemented

    task_mapper = []
    step_mapper = []
    epoch_mapper = []

    for sub_task_name, sub_task_metric in metric.items():
        task_step, task_epoch_end = task_builder[sub_task_name].validation_method(
            sub_task_metric, task=sub_task_name, preffix=preffix, ret=True
        )

        task_mapper.append(sub_task_name)
        step_mapper.append(task_step)
        epoch_mapper.append(task_epoch_end)

    def step(self, batch, batch_idx, dataloader_idx=0):
        batch['task'] = task_mapper[dataloader_idx]
        return step_mapper[dataloader_idx](self, batch, batch_idx)

    def epoch_end(self, outputs):
        metrics = []
        for idx, task_output in enumerate(outputs):
            metric = epoch_mapper[idx](self, task_output)
            metrics.append(metric)
        metric_mean = sum(metrics) / len(metrics)
        self.log(f'{task}/{preffix}_metric_mean', metric_mean, on_step=False, on_epoch=True, prog_bar=True, logger=True)
        if ret:
            return metric_mean

    return step, epoch_end


def build_method(model: Model, task_info: TaskInfo):
    multi_dataset, multi_metric = task_info.build_dataset(
        model,
        seg=model.hparams.seg_data_dir,
        pos=model.hparams.pos_data_dir,
        ner=model.hparams.ner_data_dir,
        dep=model.hparams.dep_data_dir,
        sdp=model.hparams.sdp_data_dir,
        srl=model.hparams.srl_data_dir
    )

    def train_dataloader(self):
        multi_dataloader = {
            task: torch.utils.data.DataLoader(
                task_dataset[datasets.Split.TRAIN],
                batch_size=getattr(self.hparams, f'{task}_batch_size') or self.hparams.batch_size,
                collate_fn=collate,
                num_workers=self.hparams.num_workers,
                pin_memory=True,
                shuffle=True
            )
            for task, task_dataset in multi_dataset.items()
        }
        res = MultiTaskDataloader(tau=self.hparams.tau, **multi_dataloader)
        return res

    def training_step(self, batch, batch_idx):
        result = self(**batch)
        self.log("loss", result.loss.item())
        return {"loss": result.loss}

    def val_dataloader(self):
        return [
            torch.utils.data.DataLoader(
                task_dataset[datasets.Split.VALIDATION],
                batch_size=getattr(self.hparams, f'{task}_batch_size') or self.hparams.batch_size,
                collate_fn=collate,
                num_workers=self.hparams.num_workers,
                pin_memory=True
            ) for task, task_dataset in multi_dataset.items()
        ]

    def test_dataloader(self):
        return [
            torch.utils.data.DataLoader(
                task_dataset[datasets.Split.TEST],
                batch_size=getattr(self.hparams, f'{task}_batch_size') or self.hparams.batch_size,
                collate_fn=collate,
                num_workers=self.hparams.num_workers,
                pin_memory=True
            ) for task, task_dataset in multi_dataset.items()
        ]

    # AdamW + LR scheduler
    def configure_optimizers(self: Model):
        num_epoch_steps = sum(
            (len(dataset[datasets.Split.TRAIN]) + self.hparams.batch_size - 1) // self.hparams.batch_size
            for dataset in multi_dataset.values()
        )
        num_train_steps = num_epoch_steps * self.hparams.max_epochs
        optimizer, scheduler = optimization.from_argparse_args(
            self.hparams,
            model=self,
            num_train_steps=num_train_steps,
            n_transformer_layers=self.transformer.config.num_hidden_layers
        )
        return [optimizer], [{'scheduler': scheduler, 'interval': 'step'}]

    model.configure_optimizers = types.MethodType(configure_optimizers, model)

    model.train_dataloader = types.MethodType(train_dataloader, model)
    model.training_step = types.MethodType(training_step, model)

    validation_step, validation_epoch_end = task_info.validation_method(
        multi_metric, task=task_info.task_name, preffix='val'
    )

    model.val_dataloader = types.MethodType(val_dataloader, model)
    model.validation_step = types.MethodType(validation_step, model)
    model.validation_epoch_end = types.MethodType(validation_epoch_end, model)

    test_step, test_epoch_end = task_info.validation_method(
        multi_metric, task=task_info.task_name, preffix='test'
    )

    model.test_dataloader = types.MethodType(test_dataloader, model)
    model.test_step = types.MethodType(test_step, model)
    model.test_epoch_end = types.MethodType(test_epoch_end, model)


task_info = TaskInfo(
    task_name='multitask',
    metric_name='metric_mean',
    build_dataset=build_dataset,
    validation_method=validation_method
)


def build_ner_distill_dataset(args):
    model = Model.load_from_checkpoint(
        args.resume_from_checkpoint, hparams=args
    )

    model.eval()
    model.freeze()

    dataset, metric = ner.build_dataset(
        model, args.ner_data_dir,
        ner.task_info.task_name
    )
    train_dataloader = torch.utils.data.DataLoader(
        dataset[datasets.Split.TRAIN],
        batch_size=args.batch_size,
        collate_fn=collate,
        num_workers=args.num_workers
    )

    output = os.path.join(args.ner_data_dir, ner.task_info.task_name, 'output.npz')

    if torch.cuda.is_available():
        model.cuda()
        map2cpu = lambda x: map2device(x)
        map2cuda = lambda x: map2device(x, model.device)
    else:
        map2cpu = lambda x: x
        map2cuda = lambda x: x

    with torch.no_grad():
        batchs = []
        for batch in tqdm(train_dataloader):
            batch = map2cuda(batch)
            logits = model.forward(task='ner', **batch).logits
            batch.update(logits=logits)
            batchs.append(map2cpu(batch))
        try:
            numpy.savez(
                output,
                data=convert2npy(batchs),
                extra=convert2npy({
                    'transitions': model.ner_classifier.crf.transitions,
                    'start_transitions': model.ner_classifier.crf.start_transitions,
                    'end_transitions': model.ner_classifier.crf.end_transitions
                })
            )
        except Exception as e:
            numpy.savez(output, data=convert2npy(batchs))

    print("Done")


def add_task_specific_args(parent_parser):
    parser = ArgumentParser(parents=[parent_parser], add_help=False)
    parser.add_argument('--batch_size', type=int, default=8)
    parser.add_argument('--num_workers', type=int, default=4)
    parser.add_argument('--tau', type=float, default=0.8)
    parser.add_argument('--ltp_model', type=str, default=None)
    parser.add_argument('--ltp_version', type=str, default=ltp.__version__)
    parser.add_argument('--seg_data_dir', type=str, default=None)
    parser.add_argument('--pos_data_dir', type=str, default=None)
    parser.add_argument('--ner_data_dir', type=str, default=None)
    parser.add_argument('--dep_data_dir', type=str, default=None)
    parser.add_argument('--sdp_data_dir', type=str, default=None)
    parser.add_argument('--srl_data_dir', type=str, default=None)
    parser.add_argument('--seg_batch_size', type=int, default=None)
    parser.add_argument('--pos_batch_size', type=int, default=None)
    parser.add_argument('--ner_batch_size', type=int, default=None)
    parser.add_argument('--dep_batch_size', type=int, default=None)
    parser.add_argument('--sdp_batch_size', type=int, default=None)
    parser.add_argument('--srl_batch_size', type=int, default=None)
    parser.add_argument('--build_ner_dataset', action='store_true')
    return parser


def main():
    # 如果要输出 LTP master 分支可以使用的模型，传入 ltp_adapter 参数为输出文件夹路径，如 ltp_model
    parser = ArgumentParser()

    # add task level args
    parser = add_common_specific_args(parser)
    parser = add_tune_specific_args(parser)
    parser = add_task_specific_args(parser)

    # add model specific args
    parser = Model.add_model_specific_args(parser)
    parser = optimization.add_optimizer_specific_args(parser)
    parser = Trainer.add_argparse_args(parser)

    parser.set_defaults(min_epochs=1, max_epochs=10)
    parser.set_defaults(gradient_clip_val=1.0, lr_layers_getter='get_layer_lrs_with_crf')
    args = parser.parse_args()

    if args.ltp_model is not None and args.resume_from_checkpoint is not None:
        deploy_model(args, args.ltp_version)
    elif args.build_ner_dataset:
        build_ner_distill_dataset(args)
    elif args.tune:
        from ltp.utils.common_train import tune
        tune_config = {
            # 3e-4 for Small, 1e-4 for Base, 5e-5 for Large
            "lr": tune.loguniform(args.tune_min_lr, args.tune_max_lr),

            # dataset split
            "tau": tune.choice([0.8, 0.9, 1.0]),

            # 梯度衰减
            "weight_decay": tune.choice([0.0, 0.01]),

            # 梯度裁剪
            "gradient_clip_val": tune.choice([1.0, 2.0, 3.0, 4.0, 5.0]),

            # lr scheduler
            "lr_scheduler": tune.choice([
                'linear_schedule_with_warmup',
                'polynomial_decay_schedule_with_warmup',
            ]),
        }
        tune_train(args, model_class=Model, task_info=task_info, build_method=build_method, tune_config=tune_config)
    else:
        common_train(args, model_class=Model, task_info=task_info, build_method=build_method)


if __name__ == '__main__':
    main()



from typing import Tuple

import torch
from seqeval.metrics import f1_score

from ltp.algorithms import eisner


class Metric(object):
    def step(self, batch, preds, targets, mask) -> dict:
        raise NotImplementedError()

    def epoch_end(self, step_results) -> Tuple[float, dict]:
        raise NotImplementedError()


class Acc(Metric):
    def step(self, batch, preds, targets, mask) -> dict:
        preds_true = preds[mask] == targets[mask]

        return {
            'true': torch.sum(preds_true, dtype=torch.float).item(),
            'all': preds_true.numel()
        }

    def epoch_end(self, step_results) -> Tuple[float, dict]:
        acc = sum([output[f'true'] for output in step_results]) / \
              sum([output[f'all'] for output in step_results])

        return acc, {'acc': acc}


class Seqeval(Metric):
    def __init__(self, labels: list):
        self.labels = labels

    def step(self, batch, preds, targets, mask) -> dict:
        if isinstance(preds, torch.Tensor) or isinstance(targets, torch.Tensor):
            targets[mask] = -1
            preds[mask] = -1

            targets = targets.cpu().numpy()
            preds = preds.cpu().numpy()

        labels = [[self.labels[tag] for tag in sent if tag != -1] for sent in targets]
        preds = [[self.labels[tag] for tag in sent if tag != -1] for sent in preds]

        return {'pred': preds, 'labels': labels}

    def epoch_end(self, outputs) -> Tuple[float, dict]:
        preds = sum([output['pred'] for output in outputs], [])
        labels = sum([output['labels'] for output in outputs], [])

        f1 = f1_score(y_true=labels, y_pred=preds)
        return f1, {'f1': f1}


class Graph(Metric):
    def __init__(self, punct_idx):
        self.punct_idx = punct_idx

    def get_graph_entities(self, arcs, labels):
        arcs = torch.nonzero(arcs, as_tuple=False).cpu().detach().numpy()
        labels = labels.cpu().detach().numpy()

        res = []
        for arc in arcs:
            arc = tuple(arc)
            label = labels[arc]

            res.append((*arc, label))

        return set(res)

    def step(self, batch, preds, targets, mask) -> dict:
        parc, prel = preds
        rarc, rrel = targets

        if len(parc.shape) == 3:
            parc = torch.sigmoid(parc) > 0.5
        else:
            parc = torch.argmax(parc, dim=-1)

        # cat cls
        # 去除无效位置
        mask = torch.cat([mask[:, :1], mask], dim=1)
        mask = mask.unsqueeze(1) & mask.unsqueeze(2)

        mask[:, 0] = 0
        mask.diagonal(0, 1, 2).fill_(0)  # 避免自指
        parc[~mask] = 0

        parc = parc[:, 1:, :]
        prel = torch.argmax(prel[:, 1:, :], dim=-1)

        # 对 punc 不计算分数
        punc_mask = rrel == self.punct_idx

        parc[punc_mask] = 0
        rarc[punc_mask] = 0

        predict = self.get_graph_entities(parc, prel)
        real = self.get_graph_entities(rarc, rrel)

        return {
            'correct': len(predict & real),
            'pred': len(predict),
            'true': len(real)
        }

    def epoch_end(self, outputs) -> Tuple[float, dict]:
        correct = sum([output[f'correct'] for output in outputs])
        pred = sum([output[f'pred'] for output in outputs])
        true = sum([output[f'true'] for output in outputs])

        p = correct / pred if pred > 0 else 0
        r = correct / true if true > 0 else 0
        f1 = 2 * p * r / (p + r) if (p + r > 0) else 0

        return f1, {'p': p, 'r': r, 'f1': f1}


class Tree(Metric):
    def __init__(self, punct_idx):
        self.punct_idx = punct_idx

    def step(self, batch, preds, targets, mask) -> dict:
        parc, prel = preds
        rarc, rrel = targets

        parc[:, 0, 1:] = float('-inf')
        parc.diagonal(0, 1, 2).fill_(float('-inf'))
        parc = eisner(parc, torch.cat([torch.zeros_like(mask[:, :1], dtype=torch.bool), mask], dim=1))

        prel = torch.argmax(prel, dim=-1)
        prel = prel.gather(-1, parc.unsqueeze(-1)).squeeze(-1)

        # 对 punc 不计算分数
        punc_mask = rrel != self.punct_idx
        mask = mask & punc_mask

        arc_true = (parc == rarc)[mask]
        rel_true = (prel == rrel)[mask]
        union_true = arc_true & rel_true

        return {
            "arc_true": torch.sum(arc_true, dtype=torch.float).item(),
            "uni_true": torch.sum(union_true, dtype=torch.float).item(),
            "all": union_true.numel()
        }

    def epoch_end(self, outputs) -> Tuple[float, dict]:
        uas = sum([output["arc_true"] for output in outputs]) \
              / sum([output[f"all"] for output in outputs])
        las = sum([output[f"uni_true"] for output in outputs]) \
              / sum([output[f"all"] for output in outputs])

        return las, {'uas': uas, 'las': las}


class CharTree(Metric):
    # 这里假定了 app = 0
    def __init__(self, punct_idx):
        self.app_idx = 0
        self.punct_idx = punct_idx

    def get_graph_entities(self, batch_arcs, batch_labels, batch_lens):
        seg_entities = []
        arc_entities = []
        rel_entities = []
        for batch, (arcs, labels) in enumerate(zip(batch_arcs.cpu().numpy(), batch_labels.cpu().numpy())):
            sent_length = batch_lens[batch]

            start = 0
            re_heads = []
            re_labels = []

            words = []
            ranges = []
            for word_idx, (head, label) in enumerate(zip(arcs, labels)):
                if word_idx > sent_length - 1:
                    break
                ranges.append(len(words))
                if label != self.app_idx or word_idx == sent_length - 1:
                    words.append((batch, start, word_idx + 1))
                    re_heads.append(head)
                    re_labels.append(label)
                    start = word_idx + 1
                else:
                    continue

            seg_entities.extend(words)
            for word_idx, head in enumerate(re_heads):
                if re_labels[word_idx] == -1:
                    continue

                if head == 0:
                    arc_entities.append(('root', words[word_idx]))
                    rel_entities.append(('root', re_labels[word_idx], words[word_idx]))
                else:
                    head_word_idx = ranges[head - 1]
                    head_word_span = words[head_word_idx]

                    arc_entities.append((head_word_span, words[word_idx]))
                    rel_entities.append((head_word_span, re_labels[word_idx], words[word_idx]))

        return set(arc_entities), set(rel_entities), set(seg_entities)

    def step(self, batch, preds, targets, mask) -> dict:
        seq_lens = torch.sum(mask, dim=-1, dtype=torch.int).cpu().numpy()

        parc, prel = preds
        rarc, rrel = targets

        mask = torch.cat([torch.zeros_like(mask[:, :1], dtype=torch.bool), mask], dim=1)

        parc[:, 0, 1:] = float('-inf')
        parc.diagonal(0, 1, 2).fill_(float('-inf'))
        parc = eisner(parc, mask)

        # batch_size x max_len without root
        arange_index = torch.arange(1, mask.shape[1] + 1, dtype=torch.long, device=prel.device) \
            .unsqueeze(0) \
            .repeat(mask.shape[0], 1)
        app_masks = parc.ne(arange_index)
        app_masks = app_masks.unsqueeze(-1).unsqueeze(-1).repeat(1, 1, 1, prel.shape[-1])
        app_masks[:, :, :, 1:] = 0

        prel = prel.masked_fill(app_masks, float("-inf"))
        prel = torch.argmax(prel, dim=-1)
        prel = prel.gather(-1, parc.unsqueeze(-1)).squeeze(-1)

        parc = parc[:, 1:]
        prel = prel[:, 1:]

        punc_mask = rrel == self.punct_idx
        rrel = rrel.masked_fill(punc_mask, -1)
        prel = prel.masked_fill(punc_mask, -1)

        rarc_entities, rrel_entities, rseg_entities = self.get_graph_entities(rarc, rrel, seq_lens)
        parc_entities, prel_entities, pseg_entities = self.get_graph_entities(parc, prel, seq_lens)

        return {
            'uas_correct': len(parc_entities & rarc_entities),
            'uas_pred': len(parc_entities),
            'uas_true': len(rarc_entities),
            'las_correct': len(prel_entities & rrel_entities),
            'las_pred': len(prel_entities),
            'las_true': len(rrel_entities),
            'seg_correct': len(pseg_entities & rseg_entities),
            'seg_pred': len(pseg_entities),
            'seg_true': len(rseg_entities),
        }

    def epoch_end(self, outputs) -> Tuple[float, dict]:
        uas_f1, uas_metric = self.preffix_compute('uas', outputs)
        las_f1, las_metric = self.preffix_compute('las', outputs)
        seg_f1, seg_metric = self.preffix_compute('seg', outputs)

        las_metric.update(uas_metric)
        las_metric.update(seg_metric)
        return las_f1, las_metric

    def preffix_compute(self, preffix, outputs):
        correct = sum([output[f'{preffix}_correct'] for output in outputs])
        pred = sum([output[f'{preffix}_pred'] for output in outputs])
        true = sum([output[f'{preffix}_true'] for output in outputs])

        p = correct / pred if pred > 0 else 0
        r = correct / true if true > 0 else 0
        f1 = 2 * p * r / (p + r) if (p + r > 0) else 0

        return f1, {f'{preffix}_p': p, f'{preffix}_r': r, f'{preffix}_f1': f1}



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

import types
import torch
import datasets
import torch.utils.data

from ltp import optimization
from ltp.data.utils import collate
from ltp.utils.task_info import TaskInfo
from ltp.nn import BaseModule as Model


def default_build_method(model: Model, task_info: TaskInfo):
    dataset, metric = task_info.build_dataset(
        data_dir=model.hparams.data_dir,
        task_name=task_info.task_name,
        model=model
    )

    def train_dataloader(self: Model):
        res = torch.utils.data.DataLoader(
            dataset[datasets.Split.TRAIN],
            batch_size=self.hparams.batch_size,
            collate_fn=collate,
            num_workers=self.hparams.num_workers,
            pin_memory=True,
            shuffle=True
        )
        return res

    def training_step(self: Model, batch, batch_nb):
        result = self.forward(**batch)
        self.log("loss", result.loss.item())
        return result.loss

    def val_dataloader(self: Model):
        return torch.utils.data.DataLoader(
            dataset[datasets.Split.VALIDATION],
            batch_size=self.hparams.batch_size,
            collate_fn=collate,
            num_workers=self.hparams.num_workers,
            pin_memory=True
        )

    def test_dataloader(self: Model):
        return torch.utils.data.DataLoader(
            dataset[datasets.Split.TEST],
            batch_size=self.hparams.batch_size,
            collate_fn=collate,
            num_workers=self.hparams.num_workers,
            pin_memory=True
        )

    # AdamW + LR scheduler
    def configure_optimizers(self: Model):
        num_epoch_steps = (len(dataset[datasets.Split.TRAIN]) + self.hparams.batch_size - 1) // self.hparams.batch_size
        num_train_steps = num_epoch_steps * self.hparams.max_epochs
        optimizer, scheduler = optimization.from_argparse_args(
            self.hparams,
            model=self,
            num_train_steps=num_train_steps,
            n_transformer_layers=self.transformer.config.num_hidden_layers,
        )

        return [optimizer], [{'scheduler': scheduler, 'interval': 'step'}]

    model.configure_optimizers = types.MethodType(configure_optimizers, model)

    model.train_dataloader = types.MethodType(train_dataloader, model)
    model.training_step = types.MethodType(training_step, model)

    validation_step, validation_epoch_end = task_info.validation_method(
        metric, task=task_info.task_name, preffix='val'
    )

    model.val_dataloader = types.MethodType(val_dataloader, model)
    model.validation_step = types.MethodType(validation_step, model)
    model.validation_epoch_end = types.MethodType(validation_epoch_end, model)

    test_step, test_epoch_end = task_info.validation_method(
        metric, task=task_info.task_name, preffix='test'
    )

    model.test_dataloader = types.MethodType(test_dataloader, model)
    model.test_step = types.MethodType(test_step, model)
    model.test_epoch_end = types.MethodType(test_epoch_end, model)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

import torch
from torch._six import container_abcs


def map2device(batch, device=torch.device('cpu')):
    batch_type = type(batch)
    if isinstance(batch, torch.Tensor):
        return batch.to(device)
    elif isinstance(batch, container_abcs.Mapping):
        return {key: map2device(batch[key], device=device) for key in batch}
    elif isinstance(batch, tuple) and hasattr(batch, '_fields'):  # namedtuple
        return batch_type(*(map2device(samples, device=device) for samples in zip(*batch)))
    elif isinstance(batch, container_abcs.Sequence):
        return [map2device(it, device=device) for it in batch]
    else:
        return batch


def convert2npy(batch):
    batch_type = type(batch)
    if isinstance(batch, torch.Tensor):
        return map2device(batch).numpy()
    elif isinstance(batch, container_abcs.Mapping):
        return {key: convert2npy(batch[key]) for key in batch}
    elif isinstance(batch, tuple) and hasattr(batch, '_fields'):  # namedtuple
        return batch_type(*(convert2npy(samples) for samples in zip(*batch)))
    elif isinstance(batch, container_abcs.Sequence):
        return [convert2npy(it) for it in batch]
    else:
        return batch



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

import os
import time
import argparse

import warnings
import torch
import pytorch_lightning as pl
from pytorch_lightning import Trainer, loggers
from pytorch_lightning.callbacks import ModelCheckpoint, EarlyStopping

from ltp.utils.task_info import TaskInfo
from ltp.utils.method_builder import default_build_method

warnings.filterwarnings("ignore")


def add_common_specific_args(parent_parser):
    parser = argparse.ArgumentParser(parents=[parent_parser], add_help=False)
    parser.add_argument('--test', action='store_true')
    parser.add_argument('--offline', action='store_true')
    parser.add_argument('--project', type=str, default='ltp')
    parser.add_argument('--patience', type=int, default=5)
    parser.add_argument('--seed', type=int, default=19980524)
    return parser


def common_train(args, model_class, task_info: TaskInfo, build_method=default_build_method, model_kwargs: dict = None):
    if model_kwargs is None:
        model_kwargs = {}

    pl.seed_everything(args.seed)
    if args.test:
        try:
            model = model_class.load_from_checkpoint(
                args.resume_from_checkpoint, **model_kwargs
            )
        except AttributeError as e:
            if "_gpus_arg_default" in e.args[0]:
                from ltp.patchs import pl_1_2_patch_1_1
                patched_model_path = pl_1_2_patch_1_1(args.resume_from_checkpoint)
                model = model_class.load_from_checkpoint(
                    patched_model_path, **model_kwargs
                )
            else:
                raise e
        build_method(model, task_info)
        trainer: Trainer = Trainer.from_argparse_args(args)
        trainer.test(model)
    else:
        model = model_class(args, **model_kwargs)
        build_method(model, task_info)
        early_stop_callback = EarlyStopping(
            monitor=f'{task_info.task_name}/val_{task_info.metric_name}', min_delta=1e-5, patience=args.patience,
            verbose=False, mode='max'
        )
        checkpoint_callback = ModelCheckpoint(
            monitor=f'{task_info.task_name}/val_{task_info.metric_name}', save_top_k=1, verbose=True, mode='max',
            save_last=True
        )
        this_time = time.strftime("%m-%d_%H:%M:%S", time.localtime())
        try:
            import wandb
            logger = loggers.WandbLogger(
                project=args.project,
                save_dir='lightning_logs',
                name=f'{task_info.task_name}_{this_time}',
                offline=args.offline
            )
        except Exception:
            logger = loggers.TensorBoardLogger(
                save_dir='lightning_logs', name=f'{task_info.task_name}_{this_time}', default_hp_metric=False
            )
        trainer: Trainer = Trainer.from_argparse_args(
            args,
            logger=logger,
            callbacks=[early_stop_callback],
            checkpoint_callback=checkpoint_callback
        )
        # Ready to train with new learning rate
        trainer.fit(model)
        trainer.test()


try:
    from ray import tune
    from ray.tune import CLIReporter
    from ray.tune.schedulers import ASHAScheduler
    from ray.tune.logger import DEFAULT_LOGGERS
    from ray.tune.logger import TBXLoggerCallback, CSVLoggerCallback
    from pytorch_lightning.utilities.cloud_io import load as pl_load
    from ltp.plugins.tune import TuneReportCheckpointCallback
    from os.path import expanduser


    def add_tune_specific_args(parent_parser):
        parser = argparse.ArgumentParser(parents=[parent_parser], add_help=False)
        parser.add_argument('--tune', action='store_true')
        parser.add_argument('--tune_resume', type=str, default='all')
        parser.add_argument('--tune_keep_checkpoints_num', type=int, default=1)
        parser.add_argument('--tune_min_lr', type=float, default=1e-4)
        parser.add_argument('--tune_max_lr', type=float, default=1e-3)
        parser.add_argument('--tune_gpus_per_trial', type=float, default=1.0)
        parser.add_argument('--tune_cpus_per_trial', type=float, default=5.0)
        parser.add_argument('--tune_num_samples', type=int, default=10)
        return parser


    def tune_train_once(
            config,
            checkpoint_dir=None,
            args: argparse.Namespace = None,
            model_class: type = None,
            build_method=None,
            task_info: TaskInfo = None,
            model_kwargs: dict = None,
            resume: str = None,
            group: str = None,
            log_dir: str = None,
            **kwargs
    ):
        if resume is None:
            resume = 'all'
        args_vars = vars(args)
        args_vars.update(config)

        pl.seed_everything(args.seed)
        pl_loggers = [
            loggers.CSVLogger(save_dir=tune.get_trial_dir(), name="", version="."),
            loggers.TensorBoardLogger(save_dir=tune.get_trial_dir(), name="", version=".", default_hp_metric=False),
        ]

        try:
            import wandb
            pl_loggers.append(
                loggers.WandbLogger(
                    save_dir=log_dir or 'tune_lightning_logs',
                    project=args.project,
                    name=tune.get_trial_name(),
                    id=tune.get_trial_id(),
                    offline=args.offline,
                    group=group
                )
            )
        except Exception:
            pass

        trainer_args = dict(
            logger=pl_loggers,
            progress_bar_refresh_rate=0,
            callbacks=[
                TuneReportCheckpointCallback(
                    metrics={f'tune_{task_info.metric_name}': f'{task_info.task_name}/val_{task_info.metric_name}'},
                    filename="tune.ckpt", on="validation_end"
                )
            ]
        )
        if checkpoint_dir and resume == 'all':
            trainer_args['resume_from_checkpoint'] = os.path.join(checkpoint_dir, "tune.ckpt")

        # fix slurm trainer
        os.environ["SLURM_JOB_NAME"] = "bash"
        model = model_class(args, **model_kwargs)
        build_method(model, task_info)
        trainer: Trainer = Trainer.from_argparse_args(args, **trainer_args)
        if checkpoint_dir and resume == 'model':
            ckpt = pl_load(os.path.join(checkpoint_dir, "tune.ckpt"), map_location=lambda storage, loc: storage)
            model = model._load_model_state(ckpt)
            trainer.current_epoch = ckpt["epoch"]

        trainer.fit(model)


    def tune_train(args, model_class, task_info: TaskInfo, build_method=default_build_method,
                   model_kwargs: dict = None, tune_config=None):
        if model_kwargs is None:
            model_kwargs = {}
        this_time = time.strftime("%m-%d_%H:%M:%S", time.localtime())
        experiment_name = f'{task_info.task_name}_{this_time}'

        if tune_config is None:
            config = {
                # 3e-4 for Small, 1e-4 for Base, 5e-5 for Large
                "lr": tune.loguniform(args.tune_min_lr, args.tune_max_lr),

                # -1 for disable, 0.8 for Base/Small, 0.9 for Large
                "layerwise_lr_decay_power": tune.choice([0.8, 0.9]),

                # lr scheduler
                "lr_scheduler": tune.choice(['linear_schedule_with_warmup', 'polynomial_decay_schedule_with_warmup']),
            }
        else:
            config = tune_config
        if torch.cuda.is_available():
            resources_per_trial = {"cpu": args.tune_cpus_per_trial, "gpu": args.tune_gpus_per_trial}
        else:
            resources_per_trial = {"cpu": args.tune_cpus_per_trial}
        print("resources_per_trial", resources_per_trial)

        tune_dir = os.path.abspath('tune_lightning_logs')

        analysis = tune.run(
            tune.with_parameters(
                tune_train_once,
                args=args,
                task_info=task_info,
                model_class=model_class,
                build_method=build_method,
                model_kwargs=model_kwargs,
                resume=args.tune_resume,
                group=experiment_name,
                log_dir=tune_dir,
            ),
            mode="max",
            config=config,
            num_samples=args.tune_num_samples,
            metric=f'tune_{task_info.metric_name}',
            name=experiment_name,
            progress_reporter=CLIReporter(
                parameter_columns=list(config.keys()),
                metric_columns=[
                    "loss", f'tune_{task_info.metric_name}', "training_iteration"
                ]
            ),
            callbacks=[
                TBXLoggerCallback(),
                CSVLoggerCallback()
            ],
            resources_per_trial=resources_per_trial,
            scheduler=ASHAScheduler(
                max_t=args.max_epochs + 1,  # for test
                grace_period=args.min_epochs
            ),
            queue_trials=True,
            keep_checkpoints_num=args.tune_keep_checkpoints_num,
            checkpoint_score_attr=f'tune_{task_info.metric_name}',
            local_dir=tune_dir,
        )
        print("Best hyperparameters found were: ", analysis.best_config)
        print("Best checkpoint: ", analysis.best_checkpoint)

        args_vars = vars(args)
        args_vars.update(analysis.best_config)
        model = model_class.load_from_checkpoint(
            os.path.join(analysis.best_checkpoint, "tune.ckpt"), hparams=args, **model_kwargs
        )

        pl_loggers = [
            loggers.CSVLogger(save_dir=tune.get_trial_dir(), name="", version="."),
            loggers.TensorBoardLogger(save_dir=tune.get_trial_dir(), name="", version=".", default_hp_metric=False),
        ]

        try:
            import wandb
            pl_loggers.append(
                loggers.WandbLogger(
                    save_dir=tune_dir,
                    project=args.project,
                    name=tune.get_trial_name(),
                    id=tune.get_trial_id(),
                    offline=args.offline,
                    group=experiment_name
                )
            )
        except Exception:
            pass

        trainer: Trainer = Trainer.from_argparse_args(args, logger=pl_loggers)
        build_method(model, task_info)
        trainer.test(model)



except Exception as e:
    def tune_train(*args, **kwargs):
        print("please install ray[tune]")


    def add_tune_specific_args(parent_parser):
        parser = argparse.ArgumentParser(parents=[parent_parser], add_help=False)
        parser.add_argument('--tune', action='store_true')
        return parser



def vocab_builder(func):
    from datasets import BuilderConfig

    def func_wrapper(config: BuilderConfig, **kwargs):
        """We handle string, list and dicts in datafiles"""
        if not config.data_files:
            raise ValueError(f"At least one data file must be specified, but got data_files={config.data_files}")
        data_files = config.data_files
        if isinstance(data_files, (str, list, tuple)):
            files = data_files
            if isinstance(files, str):
                files = [files]
        else:
            files = []
            for file_list in data_files.values():
                if isinstance(file_list, str):
                    files.append(file_list)
                else:
                    files.extend(file_list)
        res = func(config.data_dir, *files, **kwargs)
        return res

    return func_wrapper



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

import logging

import os
import itertools
from collections import Counter

import datasets
from os.path import join
from dataclasses import dataclass
from ltp.data.utils import iter_blocks, vocab_builder

_TRAINING_FILE = "train.txt"
_DEV_FILE = "dev.txt"
_TEST_FILE = "test.txt"


@vocab_builder
def build_vocabs(data_dir, *files):
    counters = {'predicate': (1, Counter()), 'arguments': (slice(2, None), Counter())}

    if any([os.path.exists(os.path.join(data_dir, 'vocabs', f'{key}.txt')) for key in counters]):
        return

    if not os.path.exists(os.path.join(data_dir, 'vocabs')):
        os.makedirs(os.path.join(data_dir, 'vocabs'))

    for filename in files:
        for line_num, block in iter_blocks(filename=filename):
            values = [list(value) for value in zip(*block)]

            for name, (row, counter) in counters.items():
                current = values[row]
                if not len(current):
                    continue
                item = current[0]
                if isinstance(item, list):
                    for item in current:
                        counter.update(item)
                else:
                    counter.update(current)

    for feat, (row, counter) in counters.items():
        with open(os.path.join(data_dir, 'vocabs', f'{feat}.txt'), mode='w') as f:
            # some process
            if feat == 'predicate':
                tags = sorted(counter.keys())
                tags.remove('_')
                tags = ['_'] + tags
            elif feat == 'arguments':
                tags = sorted(counter.keys())
                tags.remove('O')
                if 'B-V' in tags:
                    tags.remove('B-V')
                    tags_backup = ['O', 'B-V']
                else:
                    tags_backup = ['O']
                tags = sorted(set([tag[2:] for tag in tags]))
                tags = [f'B-{tag}' for tag in tags] + [f'I-{tag}' for tag in tags]

                tags = tags_backup + tags
            else:
                tags = ['_']
            f.write('\n'.join(tags))


def create_feature(file=None):
    if file:
        return datasets.ClassLabel(names_file=file)
    return datasets.Value('string')


@dataclass
class SrlConfig(datasets.BuilderConfig):
    """BuilderConfig for Conll2003"""

    predicate: str = None
    arguments: str = None


# ["word", "bio"]
class Srl(datasets.GeneratorBasedBuilder):
    BUILDER_CONFIG_CLASS = SrlConfig

    @staticmethod
    def default_files(data_dir) -> dict:
        return {
            datasets.Split.TRAIN: join(data_dir, _TRAINING_FILE),
            datasets.Split.VALIDATION: join(data_dir, _DEV_FILE),
            datasets.Split.TEST: join(data_dir, _TEST_FILE),
        }

    def _info(self):
        build_vocabs(self.config)
        feats = {'predicate': self.config.predicate, 'arguments': self.config.arguments}
        for key in feats:
            if feats[key] is None:
                feats[key] = os.path.join(self.config.data_dir, 'vocabs', f'{key}.txt')
        return datasets.DatasetInfo(
            features=datasets.Features(
                {
                    "form": datasets.Sequence(datasets.Value("string")),
                    "predicate": datasets.Sequence(create_feature(feats['predicate'])),
                    "arguments": datasets.Sequence(
                        datasets.Sequence(create_feature(feats['arguments']))
                    ),
                }
            ),
            supervised_keys=None,
        )

    def _split_generators(self, dl_manager):
        """We handle string, list and dicts in datafiles"""
        if not self.config.data_files:
            raise ValueError(f"At least one data file must be specified, but got data_files={self.config.data_files}")
        data_files = dl_manager.download_and_extract(self.config.data_files)
        if isinstance(data_files, (str, list, tuple)):
            files = data_files
            if isinstance(files, str):
                files = [files]
            return [datasets.SplitGenerator(name=datasets.Split.TRAIN, gen_kwargs={"files": files})]
        splits = []
        for split_name, files in data_files.items():
            if isinstance(files, str):
                files = [files]
            splits.append(datasets.SplitGenerator(name=split_name, gen_kwargs={"files": files}))
        return splits

    def _generate_examples(self, files):
        for filename in files:
            logging.info("⏳ Generating examples from = %s", filename)
            for line_num, block in iter_blocks(filename=filename):
                # last example
                words, predicate, *roles = [list(value) for value in zip(*block)]

                yield line_num, {
                    "form": words, "predicate": predicate, "arguments": roles
                }



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from .layze_import import LazyModule
from .layze_import import fake_import_pytorch_lightning

from .task_info import TaskInfo
from .length2mask import length_to_mask
from .convertor import map2device, convert2npy
from .deploy_model import deploy_model

try:
    from seqeval.metrics.sequence_labeling import get_entities
except Exception as e:
    from .get_entities import get_entities

try:
    from .common_train import common_train, tune_train
    from .common_train import add_common_specific_args, add_tune_specific_args
except Exception as e:
    common_train = LazyModule('common_train', 'pytorch_lightning')


def dataset_cache_wrapper(extra=None, extra_builder=None):
    import os
    from transformers import AutoTokenizer
    from ltp.data.dataset import DatasetDict

    def process(func, data_dir, task_name, *args, **kwargs) -> DatasetDict:
        if 'model' in kwargs:
            model = kwargs['model']
            if 'tokenizer' not in kwargs:
                kwargs['tokenizer'] = AutoTokenizer.from_pretrained(model.hparams.transformer, use_fast=True)
            if 'max_length' not in kwargs:
                kwargs['max_length'] = model.transformer.config.max_position_embeddings
        # USER BUILD
        dataset: DatasetDict = func(data_dir, task_name, *args, **kwargs)
        return dataset

    def wrapper_maker(func):

        def func_wrapper(data_dir, task_name, *args, **kwargs) -> DatasetDict:
            cache_dir = os.path.join(data_dir, task_name)

            if 'model' in kwargs:
                model = kwargs['model']
                seed = model.hparams.seed
            else:
                seed = kwargs.get('seed', 19980524)
            seed = kwargs.get('seed', seed)

            if os.path.exists(cache_dir):
                try:
                    dataset = DatasetDict.load_from_disk(cache_dir)
                except Exception as e:
                    dataset = process(func, data_dir, task_name, *args, **kwargs)
                    dataset.save_to_disk(cache_dir)
            else:
                dataset = process(func, data_dir, task_name, *args, **kwargs)
                dataset.save_to_disk(cache_dir)

            dataset = dataset.shuffle(
                seed=seed,
                indices_cache_file_names={
                    k: d._get_cache_file_path(f"{task_name}-{k}-shuffled-index-{seed}") for k, d in
                    dataset.items()
                }
            )

            if extra_builder is not None:
                return dataset, extra_builder(dataset)

            return dataset, extra

        return func_wrapper

    return wrapper_maker



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

import warnings

try:
    from .tune_checkpoint_reporter import TuneReportCallback
    from .tune_checkpoint_reporter import TuneReportCheckpointCallback
except Exception as e:
    warnings.warn("install ray[tune] to use tune model hyper")



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
# ref: https://github.com/yzhangcs/parser

import torch
import torch.nn as nn


class Triaffine(nn.Module):
    r"""
    Triaffine layer for second-order scoring.
    This function has a tensor of weights :math:`W` and bias terms if needed.
    The score :math:`s(x, y, z)` of the vector triple :math:`(x, y, z)` is computed as :math:`x^T z^T W y`.
    Usually, :math:`x` and :math:`y` can be concatenated with bias terms.
    References:
        - Yu Zhang, Zhenghua Li and Min Zhang. 2020.
          `Efficient Second-Order TreeCRF for Neural Dependency Parsing`_.
        - Xinyu Wang, Jingxian Huang, and Kewei Tu. 2019.
          `Second-Order Semantic Dependency Parsing with End-to-End Neural Networks`_.
    Args:
        in_features (int):
            The size of the input feature.
        bias_x (bool):
            If ``True``, adds a bias term for tensor :math:`x`. Default: ``False``.
        bias_y (bool):
            If ``True``, adds a bias term for tensor :math:`y`. Default: ``False``.
    .. _Efficient Second-Order TreeCRF for Neural Dependency Parsing:
        https://www.aclweb.org/anthology/2020.acl-main.302/
    .. _Second-Order Semantic Dependency Parsing with End-to-End Neural Networks:
        https://www.aclweb.org/anthology/P19-1454/
    """

    def __init__(self, in_features, bias_x=False, bias_y=False):
        super().__init__()

        self.in_features = in_features
        self.bias_x = bias_x
        self.bias_y = bias_y
        self.weight = nn.Parameter(torch.Tensor(in_features + bias_x, in_features, in_features + bias_y))

        self.reset_parameters()

    def __repr__(self):
        s = f"n_in={self.in_features}"
        if self.bias_x:
            s += f", bias_x={self.bias_x}"
        if self.bias_y:
            s += f", bias_y={self.bias_y}"

        return f"{self.__class__.__name__}({s})"

    def reset_parameters(self):
        nn.init.zeros_(self.weight)

    def forward(self, x, y, z):
        r"""
        Args:
            x (torch.Tensor): ``[batch_size, seq_len, n_in]``.
            y (torch.Tensor): ``[batch_size, seq_len, n_in]``.
            z (torch.Tensor): ``[batch_size, seq_len, n_in]``.
        Returns:
            ~torch.Tensor:
                A scoring tensor of shape ``[batch_size, seq_len, seq_len, seq_len]``.
        """

        if self.bias_x:
            x = torch.cat((x, torch.ones_like(x[..., :1])), -1)
        if self.bias_y:
            y = torch.cat((y, torch.ones_like(y[..., :1])), -1)
        w = torch.einsum('bzk,ikj->bzij', z, self.weight)
        # [batch_size, seq_len, seq_len, seq_len]
        s = torch.einsum('bxi,bzij,byj->bzxy', x, w, y)

        return s



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

import torch
from torch._six import int_classes, string_classes, container_abcs
from torch.utils.data._utils.collate import np_str_obj_array_pattern, default_collate_err_msg_format


def collate(batch):
    r"""Puts each data field into a tensor with outer dimension batch size"""

    elem = batch[0]
    elem_type = type(elem)
    if isinstance(elem, torch.Tensor):
        try:
            out = None
            if torch.utils.data.get_worker_info() is not None:
                # If we're in a background process, concatenate directly into a
                # shared memory tensor to avoid an extra copy
                numel = sum([x.numel() for x in batch])
                storage = elem.storage()._new_shared(numel)
                out = elem.new(storage)
            return torch.stack(batch, 0, out=out)
        except Exception as e:
            return torch.nn.utils.rnn.pad_sequence(batch, batch_first=True)
    elif elem_type.__module__ == 'numpy' and elem_type.__name__ != 'str_' \
            and elem_type.__name__ != 'string_':
        elem = batch[0]
        if elem_type.__name__ == 'ndarray':
            # array of string classes and object
            if np_str_obj_array_pattern.search(elem.dtype.str) is not None:
                raise TypeError(default_collate_err_msg_format.format(elem.dtype))

            return collate([torch.as_tensor(b) for b in batch])
        elif elem.shape == ():  # scalars
            return torch.as_tensor(batch)
    elif isinstance(elem, float):
        return torch.tensor(batch, dtype=torch.float64)
    elif isinstance(elem, int_classes):
        return torch.tensor(batch)
    elif isinstance(elem, string_classes):
        return batch
    elif isinstance(elem, container_abcs.Mapping):
        return {key: collate([d[key] for d in batch]) for key in elem}
    elif isinstance(elem, tuple) and hasattr(elem, '_fields'):  # namedtuple
        return elem_type(*(collate(samples) for samples in zip(*batch)))
    elif isinstance(elem, container_abcs.Sequence):
        # check to make sure that the elements in batch have consistent size
        batch = [torch.stack(it) for it in batch]
        elem_sizes = [it.shape for it in batch]
        max_sizes = (max(sizes) for sizes in zip(*elem_sizes))
        batched = torch.zeros(len(batch), *max_sizes, dtype=batch[0].dtype)
        for idx, (elem, elem_size) in enumerate(zip(batch, elem_sizes)):
            size_1, size_2 = elem_size
            batched[idx, :size_1, :size_2] = elem
        return batched

    raise TypeError(default_collate_err_msg_format.format(elem_type))



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

import numpy as np


def cycle(iterable):
    while True:
        for x in iterable:
            yield x


class MultiTaskDataloader(object):
    def __init__(self, tau=1.0, **dataloaders):
        self.dataloaders = dataloaders

        Z = sum(pow(v, tau) for v in self.dataloader_sizes.values())
        self.tasknames, self.sampling_weights = zip(*((k, pow(v, tau) / Z) for k, v in self.dataloader_sizes.items()))
        self.dataiters = {k: cycle(v) for k, v in dataloaders.items()}

    @property
    def dataloader_sizes(self):
        if not hasattr(self, '_dataloader_sizes'):
            self._dataloader_sizes = {k: len(v) for k, v in self.dataloaders.items()}
        return self._dataloader_sizes

    def __len__(self):
        return sum(v for k, v in self.dataloader_sizes.items())

    def __iter__(self):
        for i in range(len(self)):
            taskname = np.random.choice(self.tasknames, p=self.sampling_weights)
            dataiter = self.dataiters[taskname]
            batch = next(dataiter)

            batch['task'] = taskname

            yield batch



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from types import ModuleType
from typing import List


class LazyModule(ModuleType):
    requirements: List[str]

    def __init__(self, name, requirements=None):
        super().__init__(name)

        if requirements is None:
            from ltp.const import requirements
            requirements = requirements
        self.name = name
        self.requirement = requirements

    def __getattr__(self, item):
        print(f"Need Install {' '.join(self.requirement)}!!!")
        print(f"pip install {' '.join(self.requirement)}")

        if item == '__path__':
            return []

        return LazyModule(item)


class FakePytorchLightning(ModuleType):
    requirements: List[str]

    def __init__(self, name, requirements=None):
        super().__init__(name)

        if requirements is None:
            from ltp.const import requirements
            requirements = requirements
        self.name = name
        self.requirement = requirements

    def __getattr__(self, item):
        if item == '__path__':
            return []

        return FakePytorchLightning(item)


def fake_import_pytorch_lightning():
    try:
        import pytorch_lightning
    except:
        import sys
        pytorch_lightning = FakePytorchLightning('pytorch_lightning')
        pytorch_lightning_utilities = FakePytorchLightning('pytorch_lightning.utilities')
        pytorch_lightning_utilities_argparse = FakePytorchLightning('pytorch_lightning.utilities.argparse')
        pytorch_lightning_utilities_argparse_utils = FakePytorchLightning('pytorch_lightning.utilities.argparse_utils')

        sys.modules['pytorch_lightning'] = pytorch_lightning
        sys.modules['pytorch_lightning.utilities'] = pytorch_lightning_utilities
        sys.modules['pytorch_lightning.utilities.argparse'] = pytorch_lightning_utilities_argparse
        sys.modules['pytorch_lightning.utilities.argparse_utils'] = pytorch_lightning_utilities_argparse_utils



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

import logging

import os
import itertools
from collections import Counter

import datasets
from os.path import join
from dataclasses import dataclass
from ltp.data.utils import iter_blocks, vocab_builder

_TRAINING_FILE = "train.bio"
_DEV_FILE = "dev.bio"
_TEST_FILE = "test.bio"


@vocab_builder
def build_vocabs(data_dir, *files):
    counter = Counter()

    if os.path.exists(os.path.join(data_dir, 'vocabs', 'bio.txt')):
        return

    if not os.path.exists(os.path.join(data_dir, 'vocabs')):
        os.makedirs(os.path.join(data_dir, 'vocabs'))

    for filename in files:
        for line_num, block in iter_blocks(filename=filename):
            values = [list(value) for value in zip(*block)]
            counter.update(values[1])

    with open(os.path.join(data_dir, 'vocabs', 'bio.txt'), mode='w') as f:
        tags = sorted(counter.keys())
        tags.remove('O')
        f.write('\n'.join(['O'] + tags))


def create_feature(file=None):
    if file:
        return datasets.ClassLabel(names_file=file)
    return datasets.Value('string')


@dataclass
class BioConfig(datasets.BuilderConfig):
    """BuilderConfig for Conll2003"""

    bio: str = None


# ["word", "bio"]
class Bio(datasets.GeneratorBasedBuilder):
    BUILDER_CONFIG_CLASS = BioConfig

    @staticmethod
    def default_files(data_dir) -> dict:
        return {
            datasets.Split.TRAIN: join(data_dir, _TRAINING_FILE),
            datasets.Split.VALIDATION: join(data_dir, _DEV_FILE),
            datasets.Split.TEST: join(data_dir, _TEST_FILE),
        }

    def _info(self):
        build_vocabs(self.config)
        feats = {'bio': self.config.bio}
        for key in feats:
            if feats[key] is None:
                feats[key] = os.path.join(self.config.data_dir, 'vocabs', f'{key}.txt')

        return datasets.DatasetInfo(
            features=datasets.Features(
                {
                    "form": datasets.Sequence(datasets.Value("string")),
                    "bio": datasets.Sequence(create_feature(feats['bio']))
                }
            ),
            supervised_keys=None,
        )

    def _split_generators(self, dl_manager):
        """We handle string, list and dicts in datafiles"""
        if not self.config.data_files:
            raise ValueError(f"At least one data file must be specified, but got data_files={self.config.data_files}")
        data_files = dl_manager.download_and_extract(self.config.data_files)
        if isinstance(data_files, (str, list, tuple)):
            files = data_files
            if isinstance(files, str):
                files = [files]
            return [datasets.SplitGenerator(name=datasets.Split.TRAIN, gen_kwargs={"files": files})]
        splits = []
        for split_name, files in data_files.items():
            if isinstance(files, str):
                files = [files]
            splits.append(datasets.SplitGenerator(name=split_name, gen_kwargs={"files": files}))
        return splits

    def _generate_examples(self, files):
        for filename in files:
            logging.info("⏳ Generating examples from = %s", files)
            for line_num, block in iter_blocks(filename=filename):
                # last example
                words, bio = [list(value) for value in zip(*block)]

                yield line_num, {
                    "form": words, "bio": bio
                }



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

import torch
from torch import Tensor


def length_to_mask(length: Tensor, max_len: int = None, dtype=None):
    """
    将 Sequence length 转换成 Mask

    >>> lens = [3, 5, 4]
    >>> length_to_mask(length)
    >>> [[1, 1, 1, 0, 0],\
        [1, 1, 1, 1, 1], \
        [1, 1, 1, 1, 0]]

    :param length: [batch,]
    :param max_len: 最大长度
    :param dtype: nn.dtype
    :return: batch * max_len : 如果 max_len is None
    :return: batch * max(length) : 如果 max_len is None
    """
    assert len(length.shape) == 1, 'Length shape should be 1 dimensional.'
    if max_len is None:
        max_len = max_len or torch.max(length)
    mask = torch.arange(max_len, device=length.device, dtype=length.dtype). \
               expand(length.shape[0], max_len) < length.unsqueeze(1)
    if dtype is not None:
        mask = mask.to(dtype)
    return mask



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from argparse import ArgumentParser

from torch import nn
from transformers import AutoModel

from ltp.nn import BaseModule
from ltp.transformer_linear import LinearClassifier, MLPClassfier
from ltp.transformer_rel_linear import RelativeTransformerLinearClassifier
from ltp.transformer_biaffine import BiaffineClassifier, dep_loss, sdp_loss
from ltp.transformer_vi import ViClassifier
from ltp.transformer_biaffine_crf import BiaffineCRFClassifier


class TransformerMultiTask(BaseModule):
    def __init__(self, hparams, config=None):
        super().__init__()
        self.save_hyperparameters(hparams)
        if config is None:
            self.transformer = AutoModel.from_pretrained(self.hparams.transformer)
        else:
            self.transformer = AutoModel.from_config(config)
        self.dropout = nn.Dropout(self.hparams.dropout)

        transformer_hidden_size = self.transformer.config.hidden_size
        max_length = self.transformer.config.max_position_embeddings

        if self.hparams.seg_num_labels:
            if self.hparams.seg_use_mlp:
                self.seg_classifier = MLPClassfier(
                    input_size=transformer_hidden_size,
                    hidden_sizes=self.hparams.seg_hidden_sizes,
                    num_labels=self.hparams.seg_num_labels,
                    dropout=self.hparams.dropout,
                    use_crf=self.hparams.seg_use_crf,
                    use_cls=self.hparams.seg_use_cls,
                    use_sep=self.hparams.seg_use_sep,
                    crf_reduction=self.hparams.seg_crf_reduction
                )
            else:
                self.seg_classifier = LinearClassifier(
                    input_size=transformer_hidden_size,
                    num_labels=self.hparams.seg_num_labels
                )
        if self.hparams.pos_num_labels:
            if self.hparams.pos_use_mlp:
                self.pos_classifier = MLPClassfier(
                    input_size=transformer_hidden_size,
                    hidden_sizes=self.hparams.pos_hidden_sizes,
                    num_labels=self.hparams.pos_num_labels,
                    dropout=self.hparams.dropout,
                    use_crf=self.hparams.pos_use_crf,
                    use_cls=self.hparams.pos_use_cls,
                    use_sep=self.hparams.pos_use_sep,
                    crf_reduction=self.hparams.pos_crf_reduction
                )
            else:
                self.pos_classifier = LinearClassifier(
                    input_size=transformer_hidden_size,
                    num_labels=self.hparams.pos_num_labels
                )
        if self.hparams.ner_num_labels:
            self.ner_classifier = RelativeTransformerLinearClassifier(
                input_size=transformer_hidden_size,
                hidden_size=self.hparams.ner_hidden_size,
                num_layers=self.hparams.ner_num_layers,
                num_heads=self.hparams.ner_num_heads,
                num_labels=self.hparams.ner_num_labels,
                dropout=self.hparams.dropout,
                max_length=max_length,
                use_crf=self.hparams.ner_use_crf,
                use_cls=self.hparams.ner_use_cls,
                use_sep=self.hparams.ner_use_sep,
                crf_reduction=self.hparams.ner_crf_reduction,
                disable_relative_transformer=self.hparams.ner_disable_relative_transformer
            )
        if self.hparams.dep_num_labels:
            self.dep_classifier = BiaffineClassifier(
                input_size=transformer_hidden_size,
                label_num=self.hparams.dep_num_labels,
                arc_hidden_size=self.hparams.dep_arc_hidden_size,
                rel_hidden_size=self.hparams.dep_rel_hidden_size,
                loss_interpolation=self.hparams.dep_loss_interpolation,
                dropout=self.hparams.dropout,
                loss_func=dep_loss,
                char_based=self.hparams.dep_char_based
            )
        if self.hparams.sdp_num_labels:
            if self.hparams.sdp_use_vi:
                self.sdp_classifier = ViClassifier(
                    input_size=transformer_hidden_size,
                    label_num=self.hparams.sdp_num_labels,
                    dropout=self.hparams.dropout,
                    lstm_num_layers=self.hparams.sdp_lstm_num_layers,
                    lstm_hidden_size=self.hparams.sdp_lstm_hidden_size,
                    bin_hidden_size=self.hparams.sdp_bin_hidden_size,
                    arc_hidden_size=self.hparams.sdp_arc_hidden_size,
                    rel_hidden_size=self.hparams.sdp_rel_hidden_size,
                    loss_interpolation=self.hparams.sdp_loss_interpolation,
                    inference=self.hparams.sdp_inference,
                    max_iter=self.hparams.sdp_max_iter,
                )
            else:
                self.sdp_classifier = BiaffineClassifier(
                    input_size=transformer_hidden_size,
                    label_num=self.hparams.sdp_num_labels,
                    arc_hidden_size=self.hparams.sdp_arc_hidden_size,
                    rel_hidden_size=self.hparams.sdp_rel_hidden_size,
                    loss_interpolation=self.hparams.sdp_loss_interpolation,
                    dropout=self.hparams.dropout,
                    loss_func=sdp_loss,
                    char_based=self.hparams.sdp_char_based
                )
        if self.hparams.srl_num_labels:
            self.srl_classifier = BiaffineCRFClassifier(
                input_size=transformer_hidden_size,
                label_num=self.hparams.srl_num_labels,
                hidden_size=self.hparams.srl_hidden_size,
                dropout=self.hparams.dropout
            )

    @staticmethod
    def add_model_specific_args(parent_parser):
        parser = ArgumentParser(parents=[parent_parser], add_help=False, conflict_handler='resolve')
        parser.add_argument('--transformer', type=str, default="hfl/chinese-electra-base-discriminator")
        parser.add_argument('--seg_num_labels', type=int, default=2)
        parser.add_argument('--seg_hidden_sizes', nargs='+', type=int)
        parser.add_argument('--seg_use_mlp', action='store_true')
        parser.add_argument('--seg_use_crf', action='store_true')
        parser.add_argument('--seg_use_cls', action='store_true')
        parser.add_argument('--seg_use_sep', action='store_true')
        parser.add_argument('--seg_crf_reduction', type=str, default='sum')

        parser.add_argument('--pos_num_labels', type=int, default=27)
        parser.add_argument('--pos_hidden_sizes', nargs='+', type=int)
        parser.add_argument('--pos_use_mlp', action='store_true')
        parser.add_argument('--pos_use_crf', action='store_true')
        parser.add_argument('--pos_use_cls', action='store_true')
        parser.add_argument('--pos_use_sep', action='store_true')
        parser.add_argument('--pos_crf_reduction', type=str, default='sum')

        parser.add_argument('--ner_num_labels', type=int, default=13)
        parser.add_argument('--ner_use_crf', action='store_true')
        parser.add_argument('--ner_use_cls', action='store_true')
        parser.add_argument('--ner_use_sep', action='store_true')
        parser.add_argument('--ner_disable_relative_transformer', action='store_true')
        parser.add_argument('--ner_crf_reduction', type=str, default='sum')
        parser.add_argument('--ner_num_heads', type=int, default=4)
        parser.add_argument('--ner_num_layers', type=int, default=2)
        parser.add_argument('--ner_hidden_size', type=int, default=256)

        parser.add_argument('--dep_char_based', action='store_true')
        parser.add_argument('--dep_num_labels', type=int, default=14)
        parser.add_argument('--dep_arc_hidden_size', type=int, default=500)
        parser.add_argument('--dep_rel_hidden_size', type=int, default=100)
        parser.add_argument('--dep_loss_interpolation', type=float, default=0.4)

        parser.add_argument('--sdp_use_vi', action='store_true')
        parser.add_argument('--sdp_char_based', action='store_true')
        parser.add_argument('--sdp_inference', type=str, default='mfvi')
        parser.add_argument('--sdp_max_iter', type=int, default=3)
        parser.add_argument('--sdp_num_labels', type=int, default=56)
        parser.add_argument('--sdp_lstm_num_layers', type=int, default=0)
        parser.add_argument('--sdp_lstm_hidden_size', type=int, default=600)
        parser.add_argument('--sdp_arc_hidden_size', type=int, default=600)
        parser.add_argument('--sdp_bin_hidden_size', type=int, default=150)
        parser.add_argument('--sdp_rel_hidden_size', type=int, default=600)
        parser.add_argument('--sdp_loss_interpolation', type=float, default=0.4)

        parser.add_argument('--srl_num_labels', type=int, default=97)
        parser.add_argument('--srl_hidden_size', type=int, default=600)

        parser.add_argument('--dropout', type=float, default=0.1)
        return parser

    def forward(
            self,
            task,
            input_ids=None,
            attention_mask=None,
            word_index=None,
            word_attention_mask=None,
            token_type_ids=None,
            position_ids=None,
            head_mask=None,
            inputs_embeds=None,
            **kwargs
    ):
        hidden_states = self.transformer(
            input_ids,
            attention_mask,
            token_type_ids,
            position_ids,
            head_mask,
            inputs_embeds,
            output_attentions=False,
            output_hidden_states=False,
            return_dict=False,
        )
        sequence_output = hidden_states[0]
        sequence_output = self.dropout(sequence_output)

        if task == 'seg':
            return self.seg_classifier(
                input=sequence_output,
                attention_mask=attention_mask,
                word_index=word_index,
                word_attention_mask=word_attention_mask,
                **kwargs
            )
        elif task == 'pos':
            return self.pos_classifier(
                input=sequence_output,
                attention_mask=attention_mask,
                word_index=word_index,
                word_attention_mask=word_attention_mask,
                **kwargs
            )
        elif task == 'ner':
            return self.ner_classifier(
                input=sequence_output,
                attention_mask=attention_mask,
                word_index=word_index,
                word_attention_mask=word_attention_mask,
                **kwargs
            )
        elif task == 'dep':
            return self.dep_classifier(
                input=sequence_output,
                attention_mask=attention_mask,
                word_index=word_index,
                word_attention_mask=word_attention_mask,
                **kwargs
            )
        elif task == 'sdp':
            return self.sdp_classifier(
                input=sequence_output,
                attention_mask=attention_mask,
                word_index=word_index,
                word_attention_mask=word_attention_mask,
                **kwargs
            )
        elif task == 'srl':
            return self.srl_classifier(
                input=sequence_output,
                attention_mask=attention_mask,
                word_index=word_index,
                word_attention_mask=word_attention_mask,
                **kwargs
            )



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

import os
import itertools
import regex as re
from packaging import version
from typing import Union, List
from argparse import ArgumentParser

import torch
import transformers
from transformers import AutoTokenizer, AutoConfig
from transformers import cached_path, TensorType, BatchEncoding
from transformers.file_utils import is_remote_url

transformers_version = version.parse(transformers.__version__)

from ltp.algorithms import Trie, eisner, split_sentence
from ltp.transformer_multitask import TransformerMultiTask as Model
from ltp.utils import length_to_mask, get_entities, fake_import_pytorch_lightning
from ltp.patchs import model_patch_4_1_3

try:
    from torch.hub import _get_torch_home

    torch_cache_home = _get_torch_home()
except ImportError:
    torch_cache_home = os.path.expanduser(
        os.getenv("TORCH_HOME", os.path.join(os.getenv("XDG_CACHE_HOME", "~/.cache"), "torch"))
    )

default_cache_path = os.path.join(torch_cache_home, "ltp")
LTP_CACHE = os.getenv("LTP_CACHE", default_cache_path)

model_map = {
    'base': 'http://39.96.43.154/ltp/v3/base.tgz',
    'base1': 'http://39.96.43.154/ltp/v3/base1.tgz',
    'base2': 'http://39.96.43.154/ltp/v3/base2.tgz',
    'small': 'http://39.96.43.154/ltp/v3/small.tgz',
    'tiny': 'http://39.96.43.154/ltp/v3/tiny.tgz',
    'GSD': 'http://39.96.43.154/ltp/ud/gsd.tgz',
    'GSD+CRF': 'http://39.96.43.154/ltp/ud/gsd_crf.tgz',
    'GSDSimp': 'http://39.96.43.154/ltp/ud/gsd.tgz',
    'GSDSimp+CRF': 'http://39.96.43.154/ltp/ud/gsd_crf.tgz',
}


def no_gard(func):
    def wrapper(*args, **kwargs):
        with torch.no_grad():
            return func(*args, **kwargs)

    return wrapper


WORD_START = 'B-W'
WORD_MIDDLE = 'I-W'


def get_entities_with_list(labels_, itos):
    res = []
    for labels in labels_:
        labels = [itos[label] for label in labels]
        labels = get_entities(labels)
        res.append(labels)
    return res


def get_graph_entities(arcs, labels, itos):
    sequence_num = labels.size(0)
    arcs = torch.nonzero(arcs, as_tuple=False).cpu().detach().numpy().tolist()
    labels = labels.cpu().detach().numpy()

    res = [[] for _ in range(sequence_num)]
    for idx, arc_s, arc_e in arcs:
        label = labels[idx, arc_s, arc_e]
        res[idx].append((arc_s, arc_e, itos[label]))

    return res


def convert_idx_to_name(y, array_len, id2label):
    if id2label:
        return [[id2label[idx] for idx in row[:row_len]] for row, row_len in zip(y, array_len)]
    else:
        return [[idx for idx in row[:row_len]] for row, row_len in zip(y, array_len)]


class LTP(object):
    model: Model
    seg_vocab: List[str]
    pos_vocab: List[str]
    ner_vocab: List[str]
    dep_vocab: List[str]
    sdp_vocab: List[str]
    srl_vocab: List[str]

    tensor: TensorType = TensorType.PYTORCH

    def __init__(self, path: str = 'small', device=None, **kwargs):
        if device is not None:
            if isinstance(device, torch.device):
                self.device = device
            elif isinstance(device, str):
                self.device = torch.device(device)
        elif torch.cuda.is_available():
            self.device = torch.device('cuda')
        else:
            self.device = torch.device('cpu')

        if path in model_map or is_remote_url(path) or os.path.isfile(path):
            proxies = kwargs.pop("proxies", None)
            cache_dir = kwargs.pop("cache_dir", LTP_CACHE)
            force_download = kwargs.pop("force_download", False)
            resume_download = kwargs.pop("resume_download", False)
            local_files_only = kwargs.pop("local_files_only", False)
            path = cached_path(
                model_map.get(path, path),
                cache_dir=cache_dir,
                force_download=force_download,
                proxies=proxies,
                resume_download=resume_download,
                local_files_only=local_files_only,
                extract_compressed_file=True
            )
        elif not os.path.isdir(path):
            raise FileNotFoundError()
        try:
            ckpt = torch.load(os.path.join(path, "ltp.model"), map_location=self.device)
        except Exception as e:
            fake_import_pytorch_lightning()
            ckpt = torch.load(os.path.join(path, "ltp.model"), map_location=self.device)

        model_patch_4_1_3(ckpt)

        self.cache_dir = path
        transformer_config = ckpt['transformer_config']
        transformer_config['torchscript'] = True
        config = AutoConfig.for_model(**transformer_config)

        parser = ArgumentParser()
        parser = Model.add_model_specific_args(parser)
        model_args = parser.parse_args(args=[], namespace=ckpt['model_config'])

        self.model = Model(model_args, config=config).to(self.device)
        self.model.load_state_dict(ckpt['model'], strict=False)
        self.model.eval()

        self.seg_vocab = ckpt.get('seg', [WORD_MIDDLE, WORD_START])
        self.seg_vocab_dict = {tag: idx for idx, tag in enumerate(self.seg_vocab)}
        self.pos_vocab = ckpt.get('pos', [])
        self.ner_vocab = ckpt.get('ner', [])
        self.dep_vocab = ckpt.get('dep', [])
        self.sdp_vocab = ckpt.get('sdp', [])
        self.srl_vocab = [re.sub(r'ARG(\d)', r'A\1', tag.lstrip('ARGM-')) for tag in ckpt.get('srl', [])]
        self.tokenizer = AutoTokenizer.from_pretrained(path, config=self.model.transformer.config, use_fast=True)
        self.trie = Trie()
        self._model_version = ckpt.get('version', None)

    def __str__(self):
        return f"LTP {self.version} on {self.device} (model version: {self.model_version}) "

    def __repr__(self):
        return f"LTP {self.version} on {self.device} (model version: {self.model_version}) "

    @property
    def avaliable_models(self):
        return model_map.keys()

    @property
    def version(self):
        from ltp import __version__ as version
        return version

    @property
    def model_version(self):
        return self._model_version or 'unknown'

    @property
    def max_length(self):
        return self.model.transformer.config.max_position_embeddings

    def init_dict(self, path, max_window=None):
        self.trie.init(path, max_window)

    def add_words(self, words, max_window=None):
        self.trie.add_words(words)
        self.trie.max_window = max_window

    @staticmethod
    def sent_split(inputs: List[str], flag: str = "all", limit: int = 510):
        inputs = [split_sentence(text, flag=flag, limit=limit) for text in inputs]
        inputs = list(itertools.chain(*inputs))
        return inputs

    def seg_with_dict(self, inputs: List[str], tokenized: BatchEncoding, batch_prefix):
        # 进行正向字典匹配
        matching = []
        for source_text, encoding, preffix in zip(inputs, tokenized.encodings, batch_prefix):
            text = [source_text[start:end] for start, end in encoding.offsets[1:-1] if end != 0]
            matching_pos = self.trie.maximum_forward_matching(text, preffix)
            matching.append(matching_pos)
        return matching

    @no_gard
    def _seg(self, tokenizerd, is_preseged=False):
        input_ids = tokenizerd['input_ids'].to(self.device)
        attention_mask = tokenizerd['attention_mask'].to(self.device)
        token_type_ids = tokenizerd['token_type_ids'].to(self.device)
        length = torch.sum(attention_mask, dim=-1) - 2

        pretrained_output, *_ = self.model.transformer(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            return_dict=False
        )

        # remove [CLS] [SEP]
        word_cls = pretrained_output[:, :1]
        char_input = torch.narrow(pretrained_output, 1, 1, pretrained_output.size(1) - 2)
        if is_preseged:
            segment_output = None
        else:
            segment_output = self.model.seg_classifier.forward(char_input, is_processed=True)
            segment_output = segment_output.decoded or torch.argmax(segment_output.logits, dim=-1).cpu().numpy()
        return word_cls, char_input, segment_output, length

    @no_gard
    def seg(self, inputs: Union[List[str], List[List[str]]], truncation: bool = True, is_preseged=False):
        """
        分词

        Args:
            inputs: 句子列表
            truncation: 是否对过长的句子进行截断，如果为 False 可能会抛出异常
            is_preseged:  是否已经进行过分词

        Returns:
            words: 分词后的序列
            hidden: 用于其他任务的中间表示
        """

        if transformers_version.major >= 3 and transformers_version.major > 1:
            kwargs = {'is_split_into_words': is_preseged}
        else:
            kwargs = {'is_pretokenized': is_preseged}

        tokenized = self.tokenizer.batch_encode_plus(
            inputs, padding=True, truncation=truncation,
            return_tensors=self.tensor, max_length=self.max_length,
            **kwargs
        )
        cls, hidden, seg, lengths = self._seg(tokenized, is_preseged=is_preseged)

        batch_prefix = [[
            word_idx != encoding.words[idx - 1]
            for idx, word_idx in enumerate(encoding.words) if word_idx is not None
        ] for encoding in tokenized.encodings]

        # merge segments with maximum forward matching
        if self.trie.is_init and not is_preseged:
            matches = self.seg_with_dict(inputs, tokenized, batch_prefix)
            for sent_match, sent_seg in zip(matches, seg):
                for start, end in sent_match:
                    sent_seg[start] = self.seg_vocab_dict[WORD_START]
                    sent_seg[start + 1:end] = self.seg_vocab_dict[WORD_MIDDLE]
                    if end < len(sent_seg):
                        sent_seg[end] = self.seg_vocab_dict[WORD_START]

        if is_preseged:
            sentences = inputs
            word_length = [len(sentence) for sentence in sentences]

            word_idx = []
            for encodings in tokenized.encodings:
                sentence_word_idx = []
                for idx, (start, end) in enumerate(encodings.offsets[1:]):
                    if start == 0 and end != 0:
                        sentence_word_idx.append(idx)
                word_idx.append(torch.as_tensor(sentence_word_idx, device=self.device))
        else:
            segment_output = convert_idx_to_name(seg, lengths, self.seg_vocab)
            sentences = []
            word_idx = []
            word_length = []

            for source_text, length, encoding, seg_tag, preffix in \
                    zip(inputs, lengths, tokenized.encodings, segment_output, batch_prefix):
                offsets = encoding.offsets[1:length + 1]
                text = []
                last_offset = None
                for start, end in offsets:
                    text.append('' if last_offset == (start, end) else source_text[start:end])
                    last_offset = (start, end)

                for idx in range(1, length):
                    current_beg = offsets[idx][0]
                    forward_end = offsets[idx - 1][-1]
                    if forward_end < current_beg:
                        text[idx] = source_text[forward_end:current_beg] + text[idx]
                    if not preffix[idx]:
                        seg_tag[idx] = WORD_MIDDLE

                entities = get_entities(seg_tag)
                word_length.append(len(entities))
                sentences.append([''.join(text[entity[1]:entity[2] + 1]).strip() for entity in entities])
                word_idx.append(torch.as_tensor([entity[1] for entity in entities], device=self.device))

        word_idx = torch.nn.utils.rnn.pad_sequence(word_idx, batch_first=True)
        word_idx = word_idx.unsqueeze(-1).expand(-1, -1, hidden.shape[-1])  # 展开

        word_input = torch.gather(hidden, dim=1, index=word_idx)  # 每个word第一个char的向量

        if len(self.dep_vocab) + len(self.sdp_vocab) > 0:
            word_cls_input = torch.cat([cls, word_input], dim=1)
            word_cls_mask = length_to_mask(torch.as_tensor(word_length, device=self.device) + 1)
            word_cls_mask[:, 0] = False
        else:
            word_cls_input, word_cls_mask = None, None

        return sentences, {
            'word_cls': cls, 'word_input': word_input, 'word_length': word_length,
            'word_cls_input': word_cls_input, 'word_cls_mask': word_cls_mask
        }

    @no_gard
    def pos(self, hidden: dict):
        """
        词性标注
        Args:
            hidden: 分词时所得到的中间表示

        Returns:
            pos: 词性标注结果
        """
        if len(self.pos_vocab) == 0:
            return []
        postagger_output = self.model.pos_classifier(hidden['word_input'], is_processed=True)
        postagger_output = postagger_output.decoded or torch.argmax(postagger_output.logits, dim=-1).cpu().numpy()
        postagger_output = convert_idx_to_name(postagger_output, hidden['word_length'], self.pos_vocab)
        return postagger_output

    @no_gard
    def ner(self, hidden: dict, as_entities=True):
        """
        命名实体识别
        Args:
            hidden: 分词时所得到的中间表示
            as_entities: 是否以 Entity(Type, Start, End) 的形式返回

        Returns:
            pos: 命名实体识别结果
        """
        if len(self.ner_vocab) == 0:
            return []
        ner_output = self.model.ner_classifier.forward(
            hidden['word_input'], word_attention_mask=hidden['word_cls_mask'][:, 1:], is_processed=True
        )
        ner_output = ner_output.decoded or torch.argmax(ner_output.logits, dim=-1).cpu().numpy()
        ner_output = convert_idx_to_name(ner_output, hidden['word_length'], self.ner_vocab)
        return [get_entities(ner) for ner in ner_output] if as_entities else ner_output

    @no_gard
    def srl(self, hidden: dict, keep_empty=True):
        """
        语义角色标注
        Args:
            hidden: 分词时所得到的中间表示

        Returns:
            pos: 语义角色标注结果
        """
        if len(self.srl_vocab) == 0:
            return []
        srl_output = self.model.srl_classifier.forward(
            input=hidden['word_input'],
            word_attention_mask=hidden['word_cls_mask'][:, 1:],
            is_processed=True
        ).decoded
        srl_entities = get_entities_with_list(srl_output, self.srl_vocab)

        srl_labels_res = []
        for length in hidden['word_length']:
            srl_labels_res.append([])
            curr_srl_labels, srl_entities = srl_entities[:length], srl_entities[length:]
            srl_labels_res[-1].extend(curr_srl_labels)

        if not keep_empty:
            srl_labels_res = [
                [(idx, labels) for idx, labels in enumerate(srl_labels) if len(labels)]
                for srl_labels in srl_labels_res
            ]
        return srl_labels_res

    @no_gard
    def dep(self, hidden: dict, fast=True, as_tuple=True):
        """
        依存句法树
        Args:
            hidden: 分词时所得到的中间表示
            fast: 启用 fast 模式时，减少对结果的约束，速度更快，相应的精度会降低
            as_tuple: 返回的结果是否为 (idx, head, rel) 的格式，否则返回 heads, rels

        Returns:
            依存句法树结果
        """
        if len(self.dep_vocab) == 0:
            return []
        word_attention_mask = hidden['word_cls_mask']
        result = self.model.dep_classifier.forward(
            input=hidden['word_cls_input'],
            word_attention_mask=word_attention_mask[:, 1:],
            is_processed=True
        )
        dep_arc, dep_label = result.arc_logits, result.rel_logits
        dep_arc[:, 0, 1:] = float('-inf')
        dep_arc.diagonal(0, 1, 2).fill_(float('-inf'))
        dep_arc = dep_arc.argmax(dim=-1) if fast else eisner(dep_arc, word_attention_mask)

        dep_label = torch.argmax(dep_label, dim=-1)
        dep_label = dep_label.gather(-1, dep_arc.unsqueeze(-1)).squeeze(-1)

        dep_arc[~word_attention_mask] = -1
        dep_label[~word_attention_mask] = -1

        head_pred = [
            [item for item in arcs if item != -1]
            for arcs in dep_arc[:, 1:].cpu().numpy().tolist()
        ]
        rel_pred = [
            [self.dep_vocab[item] for item in rels if item != -1]
            for rels in dep_label[:, 1:].cpu().numpy().tolist()
        ]
        if not as_tuple:
            return head_pred, rel_pred
        return [
            [(idx + 1, head, rel) for idx, (head, rel) in enumerate(zip(heads, rels))]
            for heads, rels in zip(head_pred, rel_pred)
        ]

    @no_gard
    def sdp(self, hidden: dict, mode: str = 'graph'):
        """
        语义依存图（树）
        Args:
            hidden: 分词时所得到的中间表示
            mode: ['tree', 'graph', 'mix']

        Returns:
            语义依存图（树）结果
        """
        if len(self.sdp_vocab) == 0:
            return []

        word_attention_mask = hidden['word_cls_mask']
        result = self.model.sdp_classifier(
            input=hidden['word_cls_input'],
            word_attention_mask=word_attention_mask[:, 1:],
            is_processed=True
        )
        sdp_arc, sdp_label = result.arc_logits, result.rel_logits
        sdp_arc[:, 0, 1:] = float('-inf')
        sdp_arc.diagonal(0, 1, 2).fill_(float('-inf'))  # 避免自指
        sdp_label = torch.argmax(sdp_label, dim=-1)

        if mode == 'tree':
            # 语义依存树
            sdp_arc_idx = eisner(sdp_arc, word_attention_mask).unsqueeze_(-1).expand_as(sdp_arc)
            sdp_arc_res = torch.zeros_like(sdp_arc, dtype=torch.bool).scatter_(-1, sdp_arc_idx, True)
        elif mode == 'mix':
            # 混合解码
            sdp_arc_idx = eisner(sdp_arc, word_attention_mask).unsqueeze_(-1).expand_as(sdp_arc)
            sdp_arc_res = (sdp_arc.sigmoid_() > 0.5).scatter_(-1, sdp_arc_idx, True)
        else:
            # 语义依存图
            sdp_arc_res = torch.sigmoid_(sdp_arc) > 0.5

        sdp_arc_res[~word_attention_mask] = False
        sdp_label = get_graph_entities(sdp_arc_res, sdp_label, self.sdp_vocab)

        return sdp_label



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from collections import namedtuple
from argparse import ArgumentParser

import torch
from torch import nn
from torch.nn import functional as F
from transformers import AutoModel

from ltp.nn import BaseModule, MLP, CRF

TokenClassifierResult = namedtuple('TokenClassifierResult', ['loss', 'logits', 'decoded', 'labels'])


class LinearClassifier(nn.Linear):
    def __init__(self, input_size, num_labels):
        super().__init__(input_size, num_labels)

    def forward(self, input, attention_mask=None, word_index=None, word_attention_mask=None,
                labels=None, is_processed=False) -> TokenClassifierResult:
        if not is_processed:
            input = input[:, 1:-1, :]
            if word_attention_mask is None and attention_mask is not None:
                word_attention_mask = attention_mask[:, 2:] == 1
            if word_index is not None:
                input = torch.gather(input, dim=1, index=word_index.unsqueeze(-1).expand(-1, -1, input.size(-1)))

        logits = super(LinearClassifier, self).forward(input)

        loss = None

        if labels is not None:
            loss_fct = nn.CrossEntropyLoss()
            # Only keep active parts of the loss
            if word_attention_mask is not None:
                active_loss = word_attention_mask.view(-1)
                active_logits = logits.view(-1, self.out_features)[active_loss]
                active_labels = labels.view(-1)[active_loss]
                loss = loss_fct(active_logits, active_labels)
            else:
                loss = loss_fct(logits.view(-1, self.out_features), labels.view(-1))

        return TokenClassifierResult(loss=loss, logits=logits, decoded=None, labels=None)


class MLPClassfier(nn.Module):
    def __init__(self, input_size, hidden_sizes, num_labels, dropout,
                 use_cls=False, use_sep=False, use_crf=False, crf_reduction='sum'):
        super().__init__()
        self.use_cls = use_cls
        self.use_sep = use_sep
        self.linear = MLP([input_size, *hidden_sizes, num_labels], dropout=dropout)
        if use_crf:
            self.crf = CRF(num_labels, batch_first=True)
            self.crf_reduction = crf_reduction
        else:
            self.crf = None

    def forward(self, input, attention_mask=None, word_index=None, word_attention_mask=None,
                labels=None, is_processed=False) -> TokenClassifierResult:
        if not is_processed:
            if not self.use_cls:
                input = input[:, 1:, :]
            if not self.use_cls:
                input = input[:, :-1, :]

            if word_attention_mask is None:
                assert word_index is None
                bias = int(not self.use_cls) + int(not self.use_sep)
                word_attention_mask = attention_mask[:, bias:] == 1

            if word_index is not None:
                input = torch.gather(input, dim=1, index=word_index.unsqueeze(-1).expand(-1, -1, input.size(-1)))

        logits = self.linear.forward(input)

        loss = None

        if labels is not None:
            loss_fct = nn.CrossEntropyLoss()
            # Only keep active parts of the loss
            if word_attention_mask is not None and self.crf is not None:
                logits = F.log_softmax(logits, dim=-1)
                loss = - self.crf.forward(logits, labels, word_attention_mask, reduction=self.crf_reduction)
            elif word_attention_mask is not None:
                active_loss = word_attention_mask.view(-1)
                active_logits = logits.view(-1, self.classifier.out_features)[active_loss]
                active_labels = labels.view(-1)[active_loss]
                loss = loss_fct(active_logits, active_labels)
            else:
                loss = loss_fct(logits.view(-1, self.classifier.out_features), labels.view(-1))

        decoded = None
        if not self.training and self.crf is not None:
            decoded = self.crf.decode(emissions=logits, mask=word_attention_mask)
            if self.use_cls:
                decoded = [sent[1:] for sent in decoded]
                labels = labels[:, 1:]
            if self.use_sep:
                decoded = [sent[:-1] for sent in decoded]
                labels = labels[:, :-1]

        return TokenClassifierResult(loss=loss, logits=logits, decoded=decoded, labels=labels)


class TransformerLinear(BaseModule):
    def __init__(self, hparams, config=None):
        super().__init__()
        self.save_hyperparameters(hparams)
        if config is None:
            self.transformer = AutoModel.from_pretrained(self.hparams.transformer)
        else:
            self.transformer = AutoModel.from_config(config)
        self.dropout = nn.Dropout(self.hparams.dropout)
        transformer_hidden_size = self.transformer.config.hidden_size

        if self.hparams.use_mlp:
            self.classifier = MLPClassfier(
                input_size=transformer_hidden_size,
                hidden_sizes=self.hparams.hidden_sizes,
                num_labels=self.hparams.num_labels,
                dropout=self.hparams.dropout,
                use_cls=self.hparams.use_cls,
                use_crf=self.hparams.use_crf,
                use_sep=self.hparams.use_sep,
                crf_reduction=self.hparams.crf_reduction,
            )
        else:
            self.classifier = LinearClassifier(
                input_size=transformer_hidden_size,
                num_labels=self.hparams.num_labels
            )

    @staticmethod
    def add_model_specific_args(parent_parser):
        parser = ArgumentParser(parents=[parent_parser], add_help=False, conflict_handler='resolve')
        parser.add_argument('--transformer', type=str, default="hfl/chinese-electra-base-discriminator")
        parser.add_argument('--dropout', type=float, default=0.1)
        parser.add_argument('--hidden_sizes', nargs='+', type=int)
        parser.add_argument('--num_labels', type=int)
        parser.add_argument('--use_mlp', action='store_true')
        parser.add_argument('--use_crf', action='store_true')
        parser.add_argument('--use_cls', action='store_true')
        parser.add_argument('--use_sep', action='store_true')
        parser.add_argument('--crf_reduction', type=str, default='sum')
        return parser

    def forward(
            self,
            input_ids=None,
            attention_mask=None,
            word_index=None,
            word_attention_mask=None,
            token_type_ids=None,
            position_ids=None,
            head_mask=None,
            inputs_embeds=None,
            labels=None
    ) -> TokenClassifierResult:
        hidden_states = self.transformer(
            input_ids,
            attention_mask,
            token_type_ids,
            position_ids,
            head_mask,
            inputs_embeds,
            output_attentions=False,
            output_hidden_states=False,
            return_dict=False,
        )
        sequence_output = hidden_states[0]
        sequence_output = self.dropout(sequence_output)

        return self.classifier(
            input=sequence_output,
            attention_mask=attention_mask,
            word_attention_mask=word_attention_mask,
            labels=labels
        )



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

import os
from typing import Dict, List, Optional, Union

from torch import Tensor
from pytorch_lightning import Trainer, LightningModule

from ray import tune
from ray.tune.integration import pytorch_lightning as tune_pl


class TuneReportCallback(tune_pl.TuneReportCallback):
    def _handle(self, trainer: Trainer, pl_module: LightningModule):
        # Don't report if just doing initial validation sanity checks.
        if trainer.running_sanity_check:
            return
        if not self._metrics:
            report_dict = {
                k: v.item() if isinstance(v, Tensor) else v
                for k, v in trainer.callback_metrics.items()
            }
        else:
            report_dict = {}
            for key in self._metrics:
                if isinstance(self._metrics, dict):
                    metric = self._metrics[key]
                else:
                    metric = key
                metric_value = trainer.callback_metrics[metric]
                report_dict[key] = metric_value.item() if isinstance(metric_value, Tensor) else metric_value
        tune.report(**report_dict)


class TuneCheckpointCallback(tune_pl.TuneCallback):
    def __init__(self, filename: str = "checkpoint", on: Union[str, List[str]] = "validation_end"):
        super(TuneCheckpointCallback, self).__init__(on)
        self._filename = filename

    def _handle(self, trainer: Trainer, pl_module: LightningModule):
        if trainer.running_sanity_check:
            return
        with tune.checkpoint_dir(step=trainer.global_step) as checkpoint_dir:
            trainer.save_checkpoint(
                os.path.join(checkpoint_dir, self._filename))


class TuneReportCheckpointCallback(tune_pl.TuneCallback):
    def __init__(self,
                 metrics: Union[None, str, List[str], Dict[str, str]] = None,
                 filename: str = "checkpoint", on=None):
        super(TuneReportCheckpointCallback, self).__init__(on)
        if on is None:
            on = ["validation_end", "test_end"]
        self._checkpoint = TuneCheckpointCallback(filename, on)
        self._report = TuneReportCallback(metrics, on)

    def _handle(self, trainer: Trainer, pl_module: LightningModule):
        self._checkpoint._handle(trainer, pl_module)
        self._report._handle(trainer, pl_module)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from collections import namedtuple
from argparse import ArgumentParser

import torch
from torch import nn
from torch.nn import functional as F
from transformers import AutoModel

from ltp.nn import BaseModule, MLP, Bilinear, CRF

SRLResult = namedtuple('SRLResult', ['loss', 'rel_logits', 'decoded', 'labels'])


class BiaffineCRFClassifier(nn.Module):
    def __init__(self, input_size, label_num, dropout, hidden_size=300):
        super().__init__()
        self.label_num = label_num

        self.mlp_rel_h = MLP([input_size, hidden_size], output_dropout=dropout, output_activation=nn.ReLU)
        self.mlp_rel_d = MLP([input_size, hidden_size], output_dropout=dropout, output_activation=nn.ReLU)

        self.rel_atten = Bilinear(hidden_size, hidden_size, label_num, bias_x=True, bias_y=True, expand=True)
        self.crf = CRF(label_num)

    def rel_forword(self, input):
        rel_h = self.mlp_rel_h(input)
        rel_d = self.mlp_rel_d(input)

        s_rel = self.rel_atten(rel_h, rel_d).permute(0, 2, 3, 1)
        return s_rel

    def forward(self, input, attention_mask=None, word_index=None, word_attention_mask=None, labels=None,
                is_processed=False):
        if not is_processed:
            input = input[:, 1:-1, :]
            if word_attention_mask is None:
                assert word_index is None
                word_attention_mask = attention_mask[:, 2:] == 1
            if word_index is not None:
                input = torch.gather(input, dim=1, index=word_index.unsqueeze(-1).expand(-1, -1, input.size(-1)))

        s_rel = self.rel_forword(input)

        mask = word_attention_mask.unsqueeze(-1).expand(-1, -1, word_attention_mask.size(1))
        mask = mask & torch.transpose(mask, -1, -2)
        mask = mask.flatten(end_dim=1)
        index = mask[:, 0]

        mask = mask[index]
        s_rel = s_rel.flatten(end_dim=1)[index]
        crf_rel = F.log_softmax(s_rel, dim=-1)

        loss, decoded = None, None
        if labels is not None:
            labels = labels.flatten(end_dim=1)[index]
            loss = - self.crf.forward(emissions=crf_rel, tags=labels, mask=mask)

        if not self.training:
            decoded = self.crf.decode(emissions=crf_rel, mask=mask)

        return SRLResult(
            loss=loss, rel_logits=s_rel, decoded=decoded, labels=labels
        )


class TransformerBiaffineCRF(BaseModule):
    def __init__(self, hparams, config=None):
        super().__init__()
        self.save_hyperparameters(hparams)
        if config is None:
            self.transformer = AutoModel.from_pretrained(self.hparams.transformer)
        else:
            self.transformer = AutoModel.from_config(config)
        self.dropout = nn.Dropout(self.hparams.dropout)
        hidden_size = self.transformer.config.hidden_size
        self.classifier = BiaffineCRFClassifier(
            hidden_size,
            label_num=self.hparams.num_labels,
            dropout=self.hparams.dropout,
            hidden_size=self.hparams.hidden_size
        )

    @staticmethod
    def add_model_specific_args(parent_parser):
        parser = ArgumentParser(parents=[parent_parser], add_help=False, conflict_handler='resolve')
        parser.add_argument('--transformer', type=str, default="hfl/chinese-electra-base-discriminator")
        parser.add_argument('--hidden_size', type=int, default=300)
        parser.add_argument('--loss_interpolation', type=float, default=0.4)
        parser.add_argument('--dropout', type=float, default=0.1)
        parser.add_argument('--num_labels', type=int)
        return parser

    def forward(
            self,
            input_ids=None,
            attention_mask=None,
            word_index=None,
            word_attention_mask=None,
            token_type_ids=None,
            position_ids=None,
            head_mask=None,
            inputs_embeds=None,
            labels=None
    ) -> SRLResult:
        hidden_states = self.transformer(
            input_ids,
            attention_mask,
            token_type_ids,
            position_ids,
            head_mask,
            inputs_embeds,
            output_attentions=False,
            output_hidden_states=False,
            return_dict=False,
        )
        sequence_output = hidden_states[0]
        sequence_output = self.dropout(sequence_output)

        return self.classifier(
            input=sequence_output,
            word_index=word_index,
            attention_mask=attention_mask,
            word_attention_mask=word_attention_mask,
            labels=labels
        )



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

__version__ = '4.1.4.post1'

from . import const
from . import nn, utils
from . import data, optimization

from . import transformer_linear
from . import transformer_rel_linear
from . import transformer_biaffine
from . import transformer_multitask

try:
    from . import task_segmention
    from . import task_part_of_speech
    from . import task_named_entity_recognition
    from . import task_semantic_role_labeling
    from . import task_dependency_parsing
    from . import task_semantic_dependency_parsing
    from . import multitask
    from . import multitask_distill
except Exception as e:

    task_segmention = utils.LazyModule('task_segmention')
    task_part_of_speech = utils.LazyModule('task_part_of_speech')
    task_named_entity_recognition = utils.LazyModule('task_named_entity_recognition')
    task_semantic_role_labeling = utils.LazyModule('task_semantic_role_labeling')
    task_dependency_parsing = utils.LazyModule('task_dependency_parsing')
    task_semantic_dependency_parsing = utils.LazyModule('task_semantic_dependency_parsing')
    multitask = utils.LazyModule('multitask')
    multitask_distill = utils.LazyModule('multitask_distill')

from .frontend import LTP



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from .module import BaseModule

from .dropout import SharedDropout, IndependentDropout
from .mlp import MLP
from .crf import CRF
from .lstm import LSTM
from .ffnn import FFNN
from .bilinear import Bilinear
from .triaffine import Triaffine
from .relative_transformer import RelativeTransformer
from .variational_inference import LBPSemanticDependency, MFVISemanticDependency



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from functools import partial
from torch import nn
from torch.nn.modules import activation


class FFNN(nn.Sequential):
    def __init__(self, input_size, output_size, hidden_size=None, dropout=None, depth=1, last_dropout=False, **kwargs):
        activation_func = activation.LeakyReLU
        if hidden_size is None:
            hidden_size = output_size
        for activation_name, activation_args in kwargs.items():
            activation_func = partial(getattr(activation, activation_name), **activation_args)

        module = []
        for i in range(depth - 1):
            module.extend(
                [
                    nn.Linear(input_size if i == 0 else hidden_size, hidden_size),
                    activation_func()
                ] + ([] if dropout is None else [nn.Dropout(p=dropout, inplace=True)])
            )
        module.append(nn.Linear(hidden_size if len(module) > 0 else input_size, output_size))
        if last_dropout:
            module.append(nn.Dropout(p=dropout, inplace=True))
        super().__init__(*module)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from collections import namedtuple
from argparse import ArgumentParser

import torch
from torch import nn
from transformers import AutoModel, AutoConfig

from ltp.nn import MLP, Bilinear, BaseModule

GraphResult = namedtuple('GraphResult', ['loss', 'src_arc_logits', 'arc_logits', 'rel_logits'])


def dep_loss(model, s_arc, s_rel, head, labels, mask):
    head_loss = nn.CrossEntropyLoss()
    rel_loss = nn.CrossEntropyLoss()

    # ignore the first token of each sentence
    s_arc = s_arc[:, 1:, :]
    s_rel = s_rel[:, 1:, :]

    # Only keep active parts of the loss
    active_heads = head[mask]
    active_labels = labels[mask]
    s_arc, s_rel = s_arc[mask], s_rel[mask]

    s_rel = s_rel[torch.arange(len(active_heads)), active_heads]

    arc_loss = head_loss(s_arc, active_heads)
    rel_loss = rel_loss(s_rel, active_labels)
    loss = 2 * ((1 - model.loss_interpolation) * arc_loss + model.loss_interpolation * rel_loss)

    return loss


def sdp_loss(model, s_arc, s_rel, head, labels, mask):
    head_loss = nn.BCEWithLogitsLoss()
    rel_loss = nn.CrossEntropyLoss()

    # ignore the first token of each sentence
    s_arc = s_arc[:, 1:, :]
    s_rel = s_rel[:, 1:, :]

    # mask
    mask = mask.unsqueeze(-1).expand_as(s_arc)

    arc_loss = head_loss(s_arc[mask], head[mask].float())
    rel_loss = rel_loss(s_rel[head > 0], labels[head > 0])

    loss = 2 * ((1 - model.loss_interpolation) * arc_loss + model.loss_interpolation * rel_loss)

    return loss


class BiaffineClassifier(nn.Module):
    def __init__(self, input_size, label_num, dropout, arc_hidden_size=500, rel_hidden_size=100,
                 loss_interpolation=0.4, loss_func=dep_loss, char_based=False):
        super().__init__()
        self.char_based = char_based
        self.label_num = label_num
        self.loss_interpolation = loss_interpolation

        self.mlp_arc_h = MLP([input_size, arc_hidden_size], output_dropout=dropout, output_activation=nn.ReLU)
        self.mlp_arc_d = MLP([input_size, arc_hidden_size], output_dropout=dropout, output_activation=nn.ReLU)
        self.mlp_rel_h = MLP([input_size, rel_hidden_size], output_dropout=dropout, output_activation=nn.ReLU)
        self.mlp_rel_d = MLP([input_size, rel_hidden_size], output_dropout=dropout, output_activation=nn.ReLU)

        self.arc_atten = Bilinear(arc_hidden_size, arc_hidden_size, 1, bias_x=True, bias_y=False, expand=True)
        self.rel_atten = Bilinear(rel_hidden_size, rel_hidden_size, label_num, bias_x=True, bias_y=True, expand=True)

        self.loss_func = loss_func

    def forward(self, input, attention_mask=None, word_index=None,
                word_attention_mask=None, head=None, labels=None, is_processed=False):
        if not is_processed:
            input = input[:, :-1, :]
            if self.char_based:
                mask = attention_mask[:, 2:] == 1
                # use bigram ?
                # bigram = torch.cat([input[:, :-1, :].unsqueeze(2), input[:, 1:, :].unsqueeze(2)], dim=2)
                # bigram = torch.mean(bigram, dim=2)
                # input = torch.cat([input[:, :1, :], bigram], dim=1)
            else:
                mask = word_attention_mask
                if word_index is not None:
                    input = torch.cat([input[:, :1, :], torch.gather(
                        input[:, 1:, :], dim=1, index=word_index.unsqueeze(-1).expand(-1, -1, input.size(-1))
                    )], dim=1)
        else:
            mask = word_attention_mask

        arc_h = self.mlp_arc_h(input)
        arc_d = self.mlp_arc_d(input)

        rel_h = self.mlp_rel_h(input)
        rel_d = self.mlp_rel_d(input)

        s_arc = self.arc_atten(arc_d, arc_h).squeeze_(1)
        s_rel = self.rel_atten(rel_d, rel_h).permute(0, 2, 3, 1)

        loss = None
        if labels is not None:
            loss = self.loss_func(self, s_arc, s_rel, head, labels, mask)

        decode_s_arc = s_arc
        if mask is not None:
            activate_word_mask = torch.cat([mask[:, :1], mask], dim=1)
            activate_word_mask = activate_word_mask.unsqueeze(-1).expand_as(s_arc)
            activate_word_mask = activate_word_mask & activate_word_mask.transpose(-1, -2)
            decode_s_arc = s_arc.masked_fill(~activate_word_mask, float('-inf'))

        return GraphResult(loss=loss, arc_logits=decode_s_arc, rel_logits=s_rel, src_arc_logits=s_arc)


class TransformerBiaffine(BaseModule):
    def __init__(self, hparams, loss_func=dep_loss, config=None):
        super().__init__()
        self.save_hyperparameters(hparams)
        if config is None:
            self.transformer = AutoModel.from_pretrained(self.hparams.transformer)
        else:
            self.transformer = AutoModel.from_config(config)
        self.dropout = nn.Dropout(self.hparams.dropout)
        hidden_size = self.transformer.config.hidden_size
        self.classifier = BiaffineClassifier(
            hidden_size,
            label_num=self.hparams.num_labels,
            dropout=self.hparams.dropout,
            arc_hidden_size=self.hparams.arc_hidden_size,
            rel_hidden_size=self.hparams.rel_hidden_size,
            loss_interpolation=self.hparams.loss_interpolation,
            loss_func=loss_func,
            char_based=self.hparams.char_based
        )

    @staticmethod
    def add_model_specific_args(parent_parser):
        parser = ArgumentParser(parents=[parent_parser], add_help=False, conflict_handler='resolve')
        parser.add_argument('--transformer', type=str, default="hfl/chinese-electra-base-discriminator")
        parser.add_argument('--arc_hidden_size', type=int, default=500)
        parser.add_argument('--rel_hidden_size', type=int, default=200)
        parser.add_argument('--loss_interpolation', type=float, default=0.4)
        parser.add_argument('--dropout', type=float, default=0.1)
        parser.add_argument('--num_labels', type=int)
        parser.add_argument('--char_based', action='store_true')
        return parser

    def forward(
            self,
            input_ids=None,
            logits_mask=None,
            attention_mask=None,
            word_index=None,
            word_attention_mask=None,
            token_type_ids=None,
            position_ids=None,
            head_mask=None,
            inputs_embeds=None,
            head=None,
            labels=None
    ) -> GraphResult:
        hidden_states = self.transformer(
            input_ids,
            attention_mask,
            token_type_ids,
            position_ids,
            head_mask,
            inputs_embeds,
            output_attentions=False,
            output_hidden_states=False,
            return_dict=False,
        )
        sequence_output = hidden_states[0]
        sequence_output = self.dropout(sequence_output)

        return self.classifier(
            input=sequence_output,
            attention_mask=attention_mask,
            word_index=word_index,
            word_attention_mask=word_attention_mask,
            head=head,
            labels=labels
        )
